# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import ../types
import std/tables
import std/strformat
import vmath except Quat
import ../platform/gl
import arr_ref
export arr_ref

# For those unfamiliar with UBOs but familiar with textures,
# here are the similartities:

# block index = uniform location
# binding point = texture bound unit
# glUniformBlockBinding = glUniform1i
# glBindBufferBase = activeTexture + bindTexture
# glBindBuffer = bindTexture (it does NOT change the bind point!
#                             it's only for e.g. glBufferData)

# type UpdateRatio = enum
#     Manually
#     PerScene
#     PerObject
#     PerCamera

proc initUBO*[T](self: UBO, renderer: RenderManager, name: string, utype: typedesc[T], count: int): UBO =
    self.renderer = renderer
    self.name = name
    self.byte_storage = newArrRef[byte](count * sizeof(T))
    self.binding_point = -1

    renderer.enqueue proc() =
        var buffer: GLuint
        glGenBuffers 1, addr buffer
        self.buffer = GPUBuffer(buffer)
        glBindBuffer(GL_UNIFORM_BUFFER, buffer)
        glBufferData(GL_UNIFORM_BUFFER, cast[GLsizeiptr](self.byte_storage.byte_len), nil, GL_DYNAMIC_DRAW)
        # glBindBuffer(GL_UNIFORM_BUFFER, 0)
        when defined(myouUseRenderdoc):
            glObjectLabel(GL_BUFFER, buffer, GLsizei(name.len), name.cstring)
    return self

template storage*[T](self: UBO, utype: typedesc[T]): ArrRef[T] = self.byte_storage.to(T)

proc len*[T](self: UBO, utype: typedesc[T]): int =
    self.byte_storage.byte_len div sizeof(T)

# NOTE: does not preserve contents
proc resize*[T](self: UBO, utype: typedesc[T], count: int) =
    if count == self.storage(T).len:
        return
    self.renderer.enqueue proc() =
        # according to spec we can just call glBufferData with the new size
        # but I'm deleting and unbinding the buffer just in case
        self.unbind()
        self.byte_storage = newArrRef[T](count).to(byte)
        var buffer: GLuint
        glGenBuffers 1, addr buffer
        self.buffer = GPUBuffer(buffer)
        glBindBuffer(GL_UNIFORM_BUFFER, buffer)
        # TODO: set nil instead of the actual buffer,
        # and avoid using it until it's updated for the first time
        glBufferData(GL_UNIFORM_BUFFER, cast[GLsizeiptr](self.byte_storage.len), self.byte_storage.toPointer, GL_DYNAMIC_DRAW)
        # glBindBuffer(GL_UNIFORM_BUFFER, 0)
        when defined(myouUseRenderdoc):
            let name = self.name & "*" # asterisk = it was resized
            glObjectLabel(GL_BUFFER, buffer, GLsizei(name.len), name.cstring)

proc newUBO*[T](renderer: RenderManager, name: string, utype: typedesc[T], count: int = 1): UBO =
    var ubo = new UBO
    return initUBO[T](ubo, renderer, name, T, count)

proc newUBO*[T](engine: MyouEngine, name: string, utype: typedesc[T], count: int): UBO =
    var ubo = new UBO
    return initUBO[T](ubo, engine.renderer, name, T, count)

proc get_index*(self: UBO, prog: GLuint): GLuint =
    return glGetUniformBlockIndex(prog, self.name.cstring)

proc check*(self: UBO, prog: GLuint, block_index: GLuint) =
    let idx = glGetUniformBlockIndex(prog, self.name.cstring)
    assert block_index == idx, &"UBO {self.name} has different indices: stored is {block_index}, actual is {idx}"
    var blockSize: GLint
    glGetActiveUniformBlockiv(prog, block_index,
        GL_UNIFORM_BLOCK_DATA_SIZE, addr blockSize)
    assert blockSize == self.byte_storage.len, &"Block {self.name} size is {blockSize} in shader but storage is {self.byte_storage.len} bytes"

proc bind_it*(self: UBO, rebind = false) =
    if self.binding_point == -1:
        # not bound, bind it
        var next_ubo = self.renderer.next_ubo
        let old_ubo = self.renderer.bound_ubos[next_ubo]
        if old_ubo != nil:
            old_ubo.binding_point = -1
        self.renderer.bound_ubos[next_ubo] = self
        self.binding_point = next_ubo
        next_ubo += 1
        if next_ubo == self.renderer.max_uniform_buffer_bindings:
            next_ubo = 0
        self.renderer.next_ubo = next_ubo
        glBindBufferBase(GL_UNIFORM_BUFFER, self.binding_point.GLuint, self.buffer.GLuint)
    elif rebind:
        glBindBufferBase(GL_UNIFORM_BUFFER, self.binding_point.GLuint, self.buffer.GLuint)

proc unbind*(self: UBO) =
    if self.binding_point != -1:
        self.renderer.bound_ubos[self.binding_point] = nil
        glBindBufferBase(GL_UNIFORM_BUFFER, self.binding_point.GLuint, 0)
        self.binding_point = -1

proc bind_all*(ubos: seq[UBO]) =
    if ubos.len == 0:
        return
    let renderer = ubos[0].renderer
    let maxubos = renderer.max_uniform_buffer_bindings
    var next_ubo = renderer.next_ubo
    # TODO: would it be better with a small bitset?
    var used: array[HARDCODED_MAXUBOS, bool]
    for ubo in ubos:
        if ubo.binding_point != -1:
            used[ubo.binding_point] = true
    for ubo in ubos:
        if ubo.binding_point == -1:
            while used[next_ubo]:
                inc(next_ubo)
                if next_ubo == maxubos:
                    next_ubo = 0
            let old_ubo = renderer.bound_ubos[next_ubo]
            if old_ubo != nil:
                old_ubo.binding_point = -1
            renderer.bound_ubos[next_ubo] = ubo
            ubo.binding_point = next_ubo
            glBindBufferBase(GL_UNIFORM_BUFFER, next_ubo.GLuint, ubo.buffer.GLuint)
            inc(next_ubo)
            if next_ubo == maxubos:
                next_ubo = 0
    renderer.next_ubo = next_ubo

proc bind_to_prog*(self: UBO, prog: GLuint, block_index: GLuint) {.inline.} =
    self.bind_it(false)
    glUniformBlockBinding(prog, block_index, self.binding_point.GLuint)
    if not defined(release):
        self.check(prog, block_index)

proc set_prog_index*(self: UBO, prog: GLuint, block_index: GLuint) {.inline.} =
    glUniformBlockBinding(prog, block_index, self.binding_point.GLuint)
    if not defined(release):
        self.check(prog, block_index)

proc update*(self: UBO) {.inline.} =
    # NOTE: if this doesn't work in webgl, set a flag
    let len = self.byte_storage.len
    if len != 0:
        glBindBuffer(GL_UNIFORM_BUFFER, self.buffer.GLuint)
        glBufferData(GL_UNIFORM_BUFFER, len.GLsizeiptr, self.byte_storage.toPointer, GL_DYNAMIC_DRAW)
        # glBindBuffer(GL_UNIFORM_BUFFER, 0)

template is_valid*(block_index: GLuint): bool =
    block_index != GL_INVALID_INDEX

proc destroy*(self: UBO) =
    self.unbind()
    self.byte_storage = nil
    self.buffer = 0.GPUBuffer

# TODO: make a function/template/macro that generates GLSL code to declare the UBO from an object

