# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.


import std/math
# import sugar
import ../math_utils/bezier
import arr_ref

type
    CurveMapPointHandleType* = enum
        Auto, Vector, Clamped

    CurveMapPoint* = object
        vec*: array[3, (float, float)]
        h1*, h2*: CurveMapPointHandleType
    
    RampInterpolationType* = enum
        Linear, Ease, BSpline, Cardinal, Constant

    RampHueInterpolationType* = enum
        Near, Far, Cw, Ccw
    
    RampColorMode* = enum
        Rgb, Hsv, Hsl

func get_line_equations(point0, point1: (float, float)): (float, float) =
    let (x0, y0) = point0
    let (x1, y1) = point1
    let dx = max(x1-x0, 0.0000001)
    let dy = y1-y0
    let slope = dy/dx
    let y = y0 - slope*x0
    return (slope, y)

func get_line_equations*(first, last: CurveMapPoint): ((float, float), (float, float)) =
    return (
        get_line_equations(first.vec[0], first.vec[1]),
        get_line_equations(last.vec[1], last.vec[2]))

func curve_to_mapping(points: seq[CurveMapPoint], size: int, starts, ends: float): ArrRef[float] =
    result = newArrRef[float](size)
    var p0, p1 = points[0]
    var p1_idx = 0
    # extrapolate first handle
    var (x2, y2) = p1.vec[0]
    var (x3, y3) = p1.vec[1]
    var x0, x1 = ((x2-x3) * 1000)+x3
    var y0, y1 = ((y2-y3) * 1000)+y3
    let mrange = ends - starts
    let px_size = mrange / size.float
    var pos = starts + px_size/2
    for i in 0 ..< size:
        if x3 <= pos:
            p0 = p1
            p1_idx += 1
            if p1_idx <= points.high:
                p1 = points[p1_idx]
                x0 = x3; y0 = y3
                (x1,y1) = p0.vec[2]
                (x2,y2) = p1.vec[0]
                (x3,y3) = p1.vec[1]
            else:
                # last point, just extrapolate handle
                p1 = p0
                (x0,y0) = p0.vec[1]
                (x1,y1) = p0.vec[2]
                x2 = ((x1-x0) * 1000)+x0 ; x3 = x2
                y2 = ((y1-y0) * 1000)+y0 ; y3 = y2
        let rr = solve_cubic_roots(pos, x0, x1, x2, x3)
        result[i] = cubic_bezier(rr.clamp(0,1), y0, y1, y2, y3)
        pos += px_size

func hypot[T: SomeFloat](p: (T, T)): T =
    return hypot(p[0], p[1])

func calchandle_curvemap(bezt, prev, next: CurveMapPoint, has_prev, has_next: bool): CurveMapPoint =

    result = bezt
    var p2 = bezt.vec[1]
    var p1, p3: (float, float)

    if not has_prev:
        p3 = next.vec[1]
        p1 = (
            2.0 * p2[0] - p3[0],
            2.0 * p2[1] - p3[1],
        )
    else:
        p1 = prev.vec[1]

    if not has_next:
        p1 = prev.vec[1]
        p3 = (
            2.0 * p2[0] - p1[0],
            2.0 * p2[1] - p1[1],
        )
    else:
        p3 = next.vec[1]

    let dvec_a = (p2[0] - p1[0], p2[1] - p1[1])
    let dvec_b = (p3[0] - p2[0], p3[1] - p2[1])

    var len_a = hypot dvec_a
    var len_b = hypot dvec_b

    if len_a == 0.0:
        len_a = 1.0

    if len_b == 0.0:
        len_b = 1.0


    if (bezt.h1 != Vector) or (bezt.h2 != Vector):
        let tvec: (float, float) = (
            dvec_b[0] / len_b + dvec_a[0] / len_a,
            dvec_b[1] / len_b + dvec_a[1] / len_a,
        )

        let len = hypot(tvec) * 2.5614
        if len != 0.0:
            if bezt.h1 != Vector:
                len_a /= len
                result.vec[0][0] = p2[0] + tvec[0] * -len_a
                result.vec[0][1] = p2[1] + tvec[1] * -len_a

                if (bezt.h1 == Clamped) and has_next and has_prev:  # keep horizontal if extrema
                    let ydiff1 = prev.vec[1][1] - bezt.vec[1][1]
                    let ydiff2 = next.vec[1][1] - bezt.vec[1][1]
                    if (ydiff1 <= 0.0 and ydiff2 <= 0.0) or (ydiff1 >= 0.0 and ydiff2 >= 0.0):
                        result.vec[0][1] = bezt.vec[1][1]
                    elif ydiff1 <= 0.0:
                        if prev.vec[1][1] > bezt.vec[0][1]:
                            result.vec[0][1] = prev.vec[1][1]
                    else:
                        if prev.vec[1][1] < bezt.vec[0][1]:
                            result.vec[0][1] = prev.vec[1][1]
            if bezt.h2 != Vector:
                len_b /= len
                result.vec[2][0] = p2[0] + tvec[0] * len_b
                result.vec[2][1] = p2[1] + tvec[1] * len_b

                if (bezt.h2 == Clamped) and has_next and has_prev: # keep horizontal if extrema
                    let ydiff1 = prev.vec[1][1] - bezt.vec[1][1]
                    let ydiff2 = next.vec[1][1] - bezt.vec[1][1]
                    if (ydiff1 <= 0.0 and ydiff2 <= 0.0) or (ydiff1 >= 0.0 and ydiff2 >= 0.0):
                        result.vec[2][1] = bezt.vec[1][1]
                    elif ydiff1 <= 0.0:
                        if next.vec[1][1] < bezt.vec[2][1]:
                            result.vec[2][1] = next.vec[1][1]
                    else:
                        if next.vec[1][1] > bezt.vec[2][1]:
                            result.vec[2][1] = next.vec[1][1]

    if bezt.h1 == Vector:
        result.vec[0][0] = p2[0] + dvec_a[0] * -1.0 / 3.0
        result.vec[0][1] = p2[1] + dvec_a[1] * -1.0 / 3.0
    if bezt.h2 == Vector:
        result.vec[2][0] = p2[0] + dvec_b[0] * 1.0 / 3.0
        result.vec[2][1] = p2[1] + dvec_b[1] * 1.0 / 3.0

proc fix_curve_handles(points: seq[CurveMapPoint]): seq[CurveMapPoint] =
    # this ensures the curve doesn't go backwards
    if points.len < 2: return points
    var p0 = points[0]
    for p in points[1 ..< ^0]:
        var p = p
        let (x0,y0) = p0.vec[1]
        let (x1,y1) = p0.vec[2]
        let (x2,y2) = p.vec[0]
        let (x3,y3) = p.vec[1]
        if x1 > x2:
            let h1x = x0-x1
            let h1y = y0-y1
            let h2x = x3-x2
            let h2y = y3-y2
            let s = (x3-x0) / (abs(h1x)+abs(h2x))
            p0.vec[2][0] = x0 - s * h1x
            p0.vec[2][1] = y0 - s * h1y
            p.vec[0][0] = x3 - s * h2x
            p.vec[0][1] = y3 - s * h2y
        result.add p0
        p0 = p
    result.add p0

func calc_all_handles_curvemap*(points: seq[CurveMapPoint]): seq[CurveMapPoint] =
    for i,p in points:
        result.add calchandle_curvemap(p, 
            points[max(i-1,0)], points[min(i+1,points.high)],
            i != 0, i != points.high)
    # Blender changes the first and last handles when they're auto
    # I'm not sure why... but I have to do it the same way so they match
    if result.len > 2:
        for i in 0 .. 1:
            # first or last point
            template p: untyped = result[(result.len-1)*i]
            if p.h2 == Auto:
                # next point
                let p2 = result[(result.len-2)*i+(1-i)]
                let (x,y) = p.vec[1]
                template handle: untyped = p.vec[2*(1-i)]
                let (xh,yh) = handle
                let orig_len = hypot(xh-x, yh-y)
                var vec = p2.vec[2*i]
                if i == 0:
                    vec[0] = max(vec[0], x)
                else:
                    vec[0] = min(vec[0], x)
                vec[0] -= x; vec[1] -= y
                let new_len = hypot vec
                if new_len > 0.000001:
                    let s = orig_len / new_len
                    handle[0] = vec[0]*s + x
                    handle[1] = vec[1]*s + y
                    template other_handle: untyped = p.vec[2*i]
                    other_handle[0] = x - vec[0]*s
                    other_handle[1] = y - vec[1]*s

    return fix_curve_handles result

func curves_to_mapping_byte_pixels*(curves: seq[seq[CurveMapPoint]], size: int, in_start, in_end, out_scale, out_offset: float): ArrRef[uint8] =
    # draw_bez curves[3]
    # mapping = curve_to_mapping curves[3], 30
    # draw_mapping mapping
    let r_mapping = curve_to_mapping(curves[0], size, in_start, in_end)
    let g_mapping = curve_to_mapping(curves[1], size, in_start, in_end)
    let b_mapping = curve_to_mapping(curves[2], size, in_start, in_end)
    let c_mapping = if curves.len == 4:
        curve_to_mapping(curves[3], size, in_start, in_end)
    else:
        newArrRefWith(size, 1.0)
    let out_scale = out_scale * 255
    let out_offset = out_offset * 255
    var pixels = newArrRef[uint8] size * 4
    var j = 0
    for i in 0 ..< size:
        pixels[j] = round(r_mapping[i] * out_scale + out_offset).clamp(0, 255).uint8
        pixels[j+1] = round(g_mapping[i] * out_scale + out_offset).clamp(0, 255).uint8
        pixels[j+2] = round(b_mapping[i] * out_scale + out_offset).clamp(0, 255).uint8
        pixels[j+3] = round(c_mapping[i] * out_scale + out_offset).clamp(0, 255).uint8
        j += 4
    return pixels

func curves_to_mapping_float_pixels*(curves: seq[seq[CurveMapPoint]], size: int, in_start, in_end, out_scale, out_offset: float): ArrRef[float32] =
    # draw_mapping mapping
    let r_mapping = curve_to_mapping(curves[0], size, in_start, in_end)
    let g_mapping = if curves.len > 1:
        curve_to_mapping(curves[1], size, in_start, in_end)
    else: newArrRefWith(size, 1.0)
    let b_mapping = if curves.len > 2:
        curve_to_mapping(curves[2], size, in_start, in_end)
    else: newArrRefWith(size, 1.0)
    let c_mapping = if curves.len > 3:
        curve_to_mapping(curves[3], size, in_start, in_end)
    else: newArrRefWith(size, 1.0)
    var pixels = newArrRef[float32] size * 4
    var j = 0
    for i in 0 ..< size:
        pixels[j] = r_mapping[i] * out_scale + out_offset
        pixels[j+1] = g_mapping[i] * out_scale + out_offset
        pixels[j+2] = b_mapping[i] * out_scale + out_offset
        pixels[j+3] = c_mapping[i] * out_scale + out_offset
        j += 4
    return pixels

func interpolation_weights(f: float, t: RampInterpolationType): array[4, float] =
    case t:
        of Linear:
            result[1] = 1.0-f
            result[2] = f
        of Ease:
            result[1] = 1.0-f
            result[2] = f
        of BSpline:
            result[0] = -0.16666666 * f*f*f + 0.5 * f*f - 0.5 * f + 0.16666666
            result[1] = 0.5 * f*f*f - f*f + 0.66666666
            result[2] = -0.5 * f*f*f + 0.5 * f*f + 0.5 * f + 0.16666666
            result[3] = 0.16666666 * f*f*f
        of Cardinal:
            result[0] = -0.71*f*f*f + 2.0*0.71*f*f - 0.71*f
            result[1] = (2.0 - 0.71)*f*f*f + (0.71 - 3.0)*f*f + 1.0
            result[2] = (0.71 - 2.0)*f*f*f + (3.0 - 2.0*0.71)*f*f + 0.71*f
            result[3] = 0.71*f*f*f - 0.71*f*f
        of Constant:
            result[1] = 1.0


# # TODO: set NEAREST interpolation in texture for constant

func make_ramp_pixels*(steps: seq[(array[4, float32], float32)], size: int, ipotype: RampInterpolationType): ArrRef[float32] =
    var p0idx, p1idx, p2idx, p3idx = 0
    var p1_pos = -1'f32
    var p2_pos = steps[0][1]
    var inv_range = 1'f32
    let px_size = 1/size
    var pos: float32 = px_size*0.5
    var pixels = newArrRef[float32](size*4)
    var j = 0
    for i in 0 ..< size:
        if pos >= p2_pos:
            p0idx = p1idx
            p1idx = p2idx
            p1_pos = p2_pos
            if p1idx != steps.high:
                p2idx += 1
                p2_pos = steps[p2idx][1]
                let dif = p2_pos-p1_pos
                inv_range = if dif != 0: 1/dif else: 1
            p3idx = min(steps.high, p2idx+1)
        let f = (pos-p1_pos) * inv_range
        let weights = interpolation_weights(f, ipotype)
        pixels[j] = steps[p0idx][0][0]*weights[0] + steps[p1idx][0][0]*weights[1] + steps[p2idx][0][0]*weights[2] + steps[p3idx][0][0]*weights[3]
        pixels[j+1] = steps[p0idx][0][1]*weights[0] + steps[p1idx][0][1]*weights[1] + steps[p2idx][0][1]*weights[2] + steps[p3idx][0][1]*weights[3]
        pixels[j+2] = steps[p0idx][0][2]*weights[0] + steps[p1idx][0][2]*weights[1] + steps[p2idx][0][2]*weights[2] + steps[p3idx][0][2]*weights[3]
        pixels[j+3] = steps[p0idx][0][3]*weights[0] + steps[p1idx][0][3]*weights[1] + steps[p2idx][0][3]*weights[2] + steps[p3idx][0][3]*weights[3]
        j += 4
        pos += px_size
    return pixels

func rgb_to_hsv(r, g, b: float32): (float32, float32, float32) =
    var h,s,v: float32
    let cmax = max(max(r, g), b)
    let cmin = min(min(r, g), b)
    let cdelta = cmax - cmin

    v = cmax
    if cmax != 0.0:
        s = cdelta / cmax
    if s == 0.0:
        h = 0.0
    else:
        let cr = (cmax - r) * 1/cdelta
        let cg = (cmax - g) * 1/cdelta
        let cb = (cmax - b) * 1/cdelta

        h = if r == cmax: cb - cg
        elif g == cmax: 2.0 + cr -  cb
        else: 4.0 + cg - cr

        h /= 6
        if h < 0:
            h += 1

    return (h,s,v)

func hsv_to_rgb(h,s,v: float32): (float32, float32, float32) =
    if s == 0:
        return (v,v,v)

    var h = h mod 1
    if h < 0: h += 1
    if h == 1:
        h = 0
    h *= 6
    let i = floor(h)
    let f = h - i
    let p = v * (1 - s)
    let q = v * (1 - (s * f))
    let t = v * (1 - (s * (1 - f)))

    return case i:
        of 0.0: (v, t, p)
        of 1.0: (q, v, p)
        of 2.0: (p, v, t)
        of 3.0: (p, q, v)
        of 4.0: (t, p, v)
        else: (v, p, q)

func rgb_to_hsl(r, g, b: float32): (float32, float32, float32) =
    var h, s, l: float32 = 0.0
    let cmax = max(max(r, g), b)
    let cmin = min(min(r, g), b)
    l = (cmax + cmin) / 2

    if cmax != cmin:
        let d = cmax - cmin
        if l > 0.5:
            s = d / (2.0-cmax-cmin)
        else:
            s = d / (cmax+cmin)
        if cmax == r:
            h = (g - b) / d
        elif cmax == g:
            h = (b - r) / d + 2
        elif cmax == b:
            h = (r - g) / d + 4
        h /= 6
        if h < 0.0:
            h += 1.0
    return (h,s,l)

func hue2rgb(p, q, t: float32): float32 =
    var t = t mod 1
    if t < 0: t += 1
    # elif t > 1: t -= 1
    if t < 1/6: return p + (q-p) * 6.0 * t
    if t < 0.5: return q
    if t < 2/3: return p + (q-p) * (2/3 - t) * 6.0
    return p

func hsl_to_rgb(h, s, l: float32): (float32, float32, float32) =
    if s == 0:
        return (l,l,l)
    let q = if l < 0.5:
        l * (1.0 + s)
    else:
        l + s - l * s
    let p = 2.0 * l - q
    return (
        hue2rgb(p, q, h + (1/3)),
        hue2rgb(p, q, h),
        hue2rgb(p, q, h - (1/3)),
    )

func sign(f: float32): float32 =
    if f < 0: -1 else: 1

func lerp(a,b,t: float32): float32 =
    a + (b-a)*t

func make_hue_ramp_pixels*(
    steps: seq[(array[4, float32], float32)],
    size: int,
    color_mode: RampColorMode,
    ipotype_hue: RampHueInterpolationType): ArrRef[float32] =
    # starting with a relatively large number to avoid negative numbers
    # which don't modulo into positive ones in nim
    var last_h = 64'f32
    var hsx_steps = newSeqOfCap[(float32, float32, float32)] steps.len
    for step in steps:
        var (h,s,x) = if color_mode == Hsv:
            rgb_to_hsv(step[0][0], step[0][1], step[0][2])
        else:
            rgb_to_hsl(step[0][0], step[0][1], step[0][2])
        h -= last_h
        case ipotype_hue:
            of Near:
                while abs(h) > 0.5: h -= sign(h)
            of Far:
                while abs(h) > 0.5: h -= sign(h)
                h -= sign(h)
            of Cw:
                while h < 0.0: h += 1.0
            of Ccw:
                while h > 0.0: h -= 1.0
        h += last_h
        hsx_steps.add (h,s,x)
        # dump (h,s,x)
        last_h = h

    var p0_index, p1_index = 0
    var p0_pos = -1'f32
    var p1_pos = steps[0][1]
    var inv_range = 1'f32
    let px_size = 1/size
    var pos: float32 = px_size*0.5
    var pixels = newArrRef[float32](size*4)
    var j = 0
    for i in 0 ..< size:
        if pos >= p1_pos:
            p0_index = p1_index
            p0_pos = p1_pos
            if p0_index != steps.high:
                p1_index += 1
                p1_pos = steps[p1_index][1]
                let dif = p1_pos-p0_pos
                inv_range = if dif != 0: 1/dif else: 1
        let f = (pos-p0_pos) * inv_range
        pixels[j] = lerp(hsx_steps[p0_index][0], hsx_steps[p1_index][0], f)
        pixels[j+1] = lerp(hsx_steps[p0_index][1], hsx_steps[p1_index][1], f)
        pixels[j+2] = lerp(hsx_steps[p0_index][2], hsx_steps[p1_index][2], f)
        pixels[j+3] = lerp(steps[p0_index][0][3], steps[p1_index][0][3], f)
        j += 4
        pos += px_size
    if color_mode == Hsv:
        var i = 0
        while i < pixels.len:
            let (r,g,b) = hsv_to_rgb(pixels[i], pixels[i+1], pixels[i+2])
            pixels[i] = r
            pixels[i+1] = g
            pixels[i+2] = b
            i += 4
    else:
        var i = 0
        while i < pixels.len:
            let (r,g,b) = hsl_to_rgb(pixels[i], pixels[i+1], pixels[i+2])
            pixels[i] = r
            pixels[i+1] = g
            pixels[i+2] = b
            i += 4
    return pixels
