# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import std/os
import std/sequtils
import std/strformat
import std/strutils
import std/tables
import tinyre
import vmath except Quat
# import sugar

import ../types
import ../attributes
import ./blend_format
import ../util
import ../objects/mesh

template vec2(p: ptr array[16, float32]): Vec2 = cast[ptr Vec2](p)[]
template vec3(p: ptr array[16, float32]): Vec3 = cast[ptr Vec3](p)[]
template vec4(p: ptr array[16, float32]): Vec4 = cast[ptr Vec4](p)[]
# template mat4(p: ptr array[16, float32]): Mat4 = cast[ptr Mat4](p)[]

type u8vec4 = array[4, uint8]
type i8vec4 = array[4, int8]


type AttrDomain = enum
    DVertex, DLoop

type AttrData = object
    name: string
    domain: AttrDomain
    dtype: DataType
    count: int8
    buffer: ArrRef[int8]

proc equals[T](a: ArrRef[T], s:seq[T], offset: int): bool =
    for i, v in a:
        if a[i] != s[offset+i]: return false
    return true

proc copy_to[T](a: ArrRef[T], s:var seq[T], offset: int) =
    for i, v in a:
        s[offset+i] = a[i]

template fill[T](s: seq[T], v: T) =
    for i in 0 ..< s.len:
        s[i] = v

type someArr[T] = ArrRef[T] | seq[T]

# This only interleaves attributes aligned to 4 bytes per vertex
# If you want to interleave smaller attributes we need another function to combine them first
proc interleave_attributes(attributes: someArr[AttrData], indices: someArr[int32]):
    (seq[int32], seq[int32], AttributeList, int16) =

    type Data = object
        domain: AttrDomain
        # both in 32 bit words
        size: int16
        offset: int16
        buffer: ArrRef[int32]
    var attrs: seq[Data]
    
    var layout: AttributeList
    var stride, stride4: int16 = 0
    var vertex_count = 0
    for i,a in attributes:
        layout.add Attribute(name: a.name, count: a.count, dtype: a.dtype,
                            offset: stride, location: cast[int16](i))
        let size = a.count * a.dtype.size
        let size4 = size div 4
        if a.buffer != nil:
            let buffer = a.buffer.to(int32)
            if a.domain == DVertex:
                if vertex_count == 0:
                    vertex_count = buffer.len div size4
                else:
                    assert vertex_count == buffer.len div size4, &"Inconsistent vertex attribute sizes {vertex_count} {buffer.len div size4} " & ($a)[0 .. 200]
            attrs.add Data(domain: a.domain, size: size4, offset: stride4, buffer: buffer)
        stride += size
        stride4 += size4
    
    # buffer where we write the candidate vertex
    # to later compare and find another that is equal
    var buf = newArrRef[int32](stride4)

    var out_vertices = newSeq[int32](vertex_count * stride4)
    # -1 means free spot, -2 means occupied without candidates
    # any other numer means there's another candidate at that index
    var index_map = newSeqWith(vertex_count, -1'i32)
    var out_indices = newSeq[int32]()

    for loop_idx, vtx_idx in indices:
        # copy all attributes
        for j,a in attrs:
            let idx = if a.domain == DVertex:
                vtx_idx
            else:
                cast[int32](loop_idx)
            for i in 0 ..< a.size:
                buf[a.offset+i] = a.buffer[idx*a.size+i]
        # find existing vertex or create a new one
        var dest_idx = vtx_idx
        let candidate = index_map[dest_idx]
        if candidate != -1:
            while true:
                # check if we go here
                if buf.equals(out_vertices, dest_idx * stride4):
                    break
                let candidate = index_map[dest_idx]
                if candidate >= 0:
                    # compare to the next one
                    dest_idx = candidate
                else:
                    # there's no next candidate, add one
                    # and point the previous one to this
                    let idx = cast[int32](index_map.len)
                    out_vertices.setLen (idx+1) * stride4
                    index_map.add -1
                    index_map[dest_idx] = idx
                    dest_idx = idx
                    break
        index_map[dest_idx] = -2
        
        out_indices.add dest_idx
        buf.copy_to(out_vertices, dest_idx * stride4)

        # echo buf.to float32
        # echo cast[seq[float32]](out_vertices)
        # echo buf.to int8

    return (out_vertices, out_indices, layout, stride)




proc calculate_normals(vertices: ArrRef[Vec3], indices, loop_sizes: ArrRef[int32]): seq[Vec3] =
    result.setLen vertices.len
    var pos = 0
    var first, prev = vec3()
    for ls in loop_sizes:
        var n_sum = vec3()
        for i in 0 ..< ls:
            var v = vertices[indices[i+pos]]
            case i:
            of 0:
                first = v
            of 1:
                prev = v - first
            else:
                v -= first
                n_sum += prev.cross(v).normalize_safe
                prev = v
        n_sum = n_sum.normalize_safe
        for i in pos ..< pos+ls:
            result[indices[i]] += n_sum
        pos += ls
    for i,v in result:
        result[i] = v.normalize_safe

proc loadMeshImpl*(self: BlendLoader, obn, mesh: FNode, ob: GameObject,
    use_weights=true, required_tangents: seq[string]) =

    var use_weights = use_weights
    var vertices = newArrRef[Vec3](mesh.totvert.i32[0])
    var kb = mesh.key.`block`.first
    if kb.isNil:
        if mesh.mvert.valid:
            for i,co in mesh.mvert[0 ..< vertices.len].co:
                vertices[i] = co.f32.vec3
    else:
        # Use basis as source of "original" vertices
        let data = kb.data.get_array(vertices.len, Vec3)
        for i in 0 ..< vertices.len:
            vertices[i] = data[i]
    var indices = newArrRef[int32](mesh.totloop.i32[0])
    var indices_pending = true
    var poly_count = mesh.totpoly.i32[0]
    if poly_count == 0:
        return
    var loop_sizes = newArrRef[int32](poly_count)
    var use_new_indices = false
    try:
        if mesh.poly_offset_indices.valid:
            use_new_indices = true
    except KeyError:
        discard
    if use_new_indices:
        # Blender >= 4.0
        var last = mesh.poly_offset_indices.i32[0]
        for i,idx in mesh.poly_offset_indices[1 .. poly_count]:
            # dump (i,idx-last)  # bad compiler error
            let s = idx.i32[0] - last
            # dump s.type
            loop_sizes[(i-1).Natural] = s
            last = idx.i32[0]
    else:
        # Blender <= 3.6
        for i,t in mesh.mpoly[0 ..< loop_sizes.len].totloop:
            loop_sizes[i] = t.i32[0]
    # if there are no materials we'll pretend there's one
    var material_count = max(1, mesh.totcol.i16[0])
    var materials = newArrRef[int16](loop_sizes.len)
    var used_materials = newSeq[bool](material_count)
    var materials_pending = true
    
    var attributes: seq[AttrData]
    var uv_names: seq[string]

    for layer in mesh.ldata.arr_len(layers, totlayer):
        # echo layer["type"].i32[0]," l ",layer.name.str
        var data = layer?.data
        if data.isNil: continue
        case layer["type"].i32[0]:
        of 16: # classic UV, deprecated
            var uv_data = newArrRef[Vec2](indices.len)
            for i,uv in data[0..<indices.len].uv:
                uv_data[i] = uv.f32.vec2
            uv_names.add layer.name.str
            attributes.add AttrData(
                domain: DLoop,
                name: "uv_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                dtype: Float, count: 2,
                buffer: uv_data.to(int8)
            )
        of 49: # float2, assuming UV
            var uv_data = newArrRef[Vec2](indices.len)
            for i,uv in data[0..<indices.len].x:
                uv_data[i] = uv.f32.vec2
            uv_names.add layer.name.str
            attributes.add AttrData(
                domain: DLoop,
                name: "uv_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                dtype: Float, count: 2,
                buffer: uv_data.to(int8)
            )
        of 17:
            var col_data = newArrRef[u8vec4](indices.len)
            for i,col in data[0..<indices.len].r:
                col_data[i] = cast[ptr u8vec4](col.i8)[]
            attributes.add AttrData(
                name: "vc_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                domain: DLoop,
                dtype: UByte, count: 4,
                buffer: col_data.to(int8)
            )
        of 47:
            var col_data = newArrRef[Vec4](indices.len)
            for i,col in data[0..<indices.len].color:
                col_data[i] = col.f32.vec4
            attributes.add AttrData(
                domain: DLoop,
                name: "vc_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                dtype: Float, count: 4,
                buffer: col_data.to(int8)
            )
        # of 26:
        #     # indices_pending = false
        #     # data.set_type "MLoop"
        #     # for i,v in data[0 ..< indices.len].v:
        #     #     indices[i] = v.i32[0]
        #     discard

        # of 41: custom normals (vec2 i16)

        of 11: # int32
            case layer.name.str:
                of ".corner_vert":
                    indices_pending = false
                    for i,v in data[0 ..< indices.len].i:
                        indices[i] = v.i32[0]
                of ".corner_edge":
                    discard
                else:
                    echo "Unknown int32 loop attribute, name ",layer.name.str
        of 50: # bool
            case layer.name.str.split(".")[1]:
                of "vs":
                    # UV map of all zeros or all ones
                    # TODO: use static value instead of ignoring
                    discard
                else:
                    echo "Unknown bool loop attribute, name ",layer.name.str
        else:
            echo "Unknown loop attribute type ", layer["type"].i32[0],", name ",layer.name.str
            # dump data
    for layer in mesh.vdata.arr_len(layers, totlayer):
        # echo layer["type"].i32[0]," v ",layer.name.str
        let data = layer?.data
        if data.isNil: continue
        case layer["type"].i32[0]:
        # of 0:
        #     echo "it has mvert"
        #     discard # same as mvert
        # of 2: discard # same as dvert
        of 47:
            var col_data = newArrRef[Vec4](vertices.len)
            for i,col in data[0..<vertices.len].color:
                col_data[i] = col.f32.vec4
            attributes.add AttrData(
                name: "vc_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                domain: DVertex,
                dtype: Float, count: 4,
                buffer: col_data.to(int8)
            )
        of 48: # float3
            case layer.name.str:
                of "position":
                    # TODO: skip if we already read them from basis key
                    #       or from blender 3.6?
                    for i,v in data[0 ..< vertices.len].x:
                        vertices[i] = v.f32.vec3
                else:
                    echo "Unknown int32 loop attribute, name ", layer.name.str
        of 50: # bool
            case layer.name.str:
                of ".select_vert":
                    discard
                else:
                    echo "Unknown bool loop attribute, name ", layer.name.str
        of 17:
            var col_data = newArrRef[u8vec4](vertices.len)
            for i,col in data[0..<vertices.len].r:
                col_data[i] = cast[ptr u8vec4](col.i8)[]
            attributes.add AttrData(
                name: "vc_" & layer.name.str.replace(re"[^_A-Za-z0-9]", ""),
                domain: DVertex,
                dtype: UByte, count: 4,
                buffer: col_data.to(int8)
            )
        else:
            echo "Unknown vertex attribute type ", layer["type"].i32[0],", name ", repr(layer.name.str)
    
    for layer in mesh.pdata.arr_len(layers, totlayer):
        # echo layer["type"].i32[0]," f ",layer.name.str
        let data = layer?.data
        if data.isNil: continue
        case layer["type"].i32[0]:

        of 11: # int32
            case layer.name.str:
                of "material_index":
                    materials_pending = false
                    for i,v in data[0 ..< poly_count]:
                        let m = clamp(v.i.i32[0], 0'i32, material_count)
                        used_materials[m] = true
                        materials[i] = m.int16
                else:
                    echo "Unknown int32 loop attribute, name ",layer.name.str
        of 50: # bool
            case layer.name.str:
                of ".select_poly":
                    discard
                of "sharp_face":
                    discard # TODO
                else:
                    echo "Unknown bool loop attribute, name ", layer.name.str
        else:
            echo "Unknown face attribute type ", layer["type"].i32[0],", name ", repr(layer.name.str)
    
    # Blender <= 3.6
    if indices_pending and mesh.mloop.valid:
        for i,v in mesh.mloop[0 ..< indices.len].v:
            indices[i] = v.i32[0]
    if materials_pending:
        # when there are several materials but only the first one is used,
        # blender just doesn't save any material attribute at all
        # (so we need to check if mpoly exists)
        if material_count > 1 and mesh.mpoly.valid:
            for i,m in mesh.mpoly[0 ..< materials.len].mat_nr:
                let m = clamp(m.i16[0], 0'i16, material_count)
                used_materials[m] = true
                materials[i] = m
        else:
            used_materials[0] = true
    
    while not used_materials[^1]:
        material_count -= 1
        used_materials.setLen material_count
    
    if use_weights:
        # TODO: vertex weights
        discard
        # TODO: check armature bones to get weights which weights to use
        # for dv in mesh.dvert[0 ..< vertices.len]:
        #     dump dv

    # TODO: shape keys

    # calculate normals
    let normals = calculate_normals(vertices, indices, loop_sizes)
    var normals_byte = newArrRef[i8vec4](normals.len)
    for i,n in normals:
        normals_byte[i] = [int8(n.x*127), int8(n.y*127), int8(n.z*127), 0]
    
    # add empty tangents (to be calculated later)
    if required_tangents.len != 0:
        var tg_attributes: seq[AttrData]
        var unused = required_tangents
        var active_tg_name = ""
        for i,uv_name in uv_names:
            let tg_name = "tg_" & uv_name.replace(re"[^_A-Za-z0-9]", "")
            # TODO: use active UV instead of first
            if i == 0:
                active_tg_name = tg_name
            if tg_name in required_tangents:
                tg_attributes.add AttrData(
                    name: tg_name, domain: DLoop,
                    dtype: Byte, count: 4,
                )
                unused.remove tg_name
        # if there are missing required tangents, ensure the active UV got them
        if unused.len != 0 and uv_names.len != 0:
            if active_tg_name notin required_tangents:
                attributes.add AttrData(
                    name: active_tg_name, domain: DLoop,
                    dtype: Byte, count: 4,
                )
        attributes = attributes & tg_attributes
    
    # # list of materials per index
    # var mat_per_index: seq[int16](indices.len)
    # mat_per_index.setLen 0
    # for i,ls in loop_sizes:
    #     for j in 0 ..< ls:
    #         mat_per_index.add materials[i]
    
    attributes = @[
        AttrData(
            name: "vertex", domain: DVertex,
            dtype: Float, count: 3,
            buffer: vertices.to(int8)
        ),
        AttrData(
            name: "vnormal", domain: DVertex,
            dtype: Byte, count: 4,
            buffer: normals_byte.to(int8)
        )
    ] & attributes
    let (va, ia, layout, stride) = interleave_attributes(attributes, indices)
    let stride4 = stride div 4

    # triangulate by material
    # todo: earcut n-gons?
    # todo: calculate normals after triangulating instead of before?
    #       triangle counts = loop_sizes - 2
    var trindices_per_material = newSeq[seq[int32]](material_count)
    template trindices: seq[int32] = trindices_per_material[mat_idx]
    var pos, fi = 0
    for ls in loop_sizes:
        let mat_idx = materials[fi]
        var first = ia[pos]
        var prev = ia[pos+1]
        for i in pos+2 ..< pos+ls:
            let j = ia[i]
            trindices.add first
            trindices.add prev
            trindices.add j
            prev = j
        pos += ls
        fi.inc
    
    let use_indices = self.use_indices
    
    # separate vertices
    # TODO: don't do this and sort vertices instead?
    var va_per_material: seq[seq[int32]]
    var ia_per_material: seq[seq[int32]]
    if use_indices:
        if material_count > 1:
            var va_map = newSeqWith(va.len, -1)
            for j,indices in trindices_per_material:
                if j != 0: va_map.fill -1
                var va_out: seq[int32]
                var ia_out: seq[int32]
                var count = 0
                for i,idx in indices:
                    var oidx = va_map[idx]
                    if oidx == -1:
                        oidx = count
                        count += 1
                        for k in idx*stride4 ..< (idx+1)*stride4:
                            va_out.add va[k]
                    ia_out.add cast[int32](oidx)
                va_per_material.add va_out
                ia_per_material.add ia_out
        else:
            # TODO: avoid concatenating a single one
            va_per_material.add va
            ia_per_material.add trindices_per_material[0]
    else:
        var va_map = newSeqWith(va.len, -1)
        for j,indices in trindices_per_material:
            if j != 0: va_map.fill -1
            var va_out: seq[int32]
            for i,idx in indices:
                for k in idx*stride4 ..< (idx+1)*stride4:
                    va_out.add va[k]
            va_per_material.add va_out

    # TODO: remove empty materials

    # TODO: 16 bit indices, looking at the size of vertex arrays

    var final_va: seq[ArrRef[float32]]
    var final_ia: seq[ArrRef[int32]]
    var index_types: seq[DataType]
    for i in 0 ..< material_count:
        final_va.add newArrRef(va_per_material[i]).to(float32)
        if use_indices:
            final_ia.add newArrRef(ia_per_material[i])
        index_types.add UInt

    ob.get_mesh.generate_tangents = required_tangents.len != 0
    ob.get_mesh.load_from_va_ia(final_va, final_ia, layout, index_types)
    # ob.get_mesh.debug_print_vertices(0,10)

    ob.get_mesh.update_bounding_box
