# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

{.hint[ClosureWithoutEnv]: off.}
{.hint[XDeclaredButNotUsed]: off.}

import std/bitops
import std/hashes
import std/math
import std/sets
# import std/sequtils
import std/strformat
import std/strutils
import std/tables
import tinyre
import arr_ref
import ../types
import ./blend_curve_maps
import ./blend_format
import ./glsl_functions
import vmath except Quat

type Expr = object
    str: string
    btype: BlendBasicType
    vlen: uint32

template `$`(e: Expr): string = e.str
template flt(s: string): Expr = Expr(str: s, btype: B_float, vlen: 1)
template vec2(s: string): Expr = Expr(str: s, btype: B_float, vlen: 2)
template vec3(s: string): Expr = Expr(str: s, btype: B_float, vlen: 3)
template vec4(s: string): Expr = Expr(str: s, btype: B_float, vlen: 4)
template mat3(s: string): Expr = Expr(str: s, btype: B_float, vlen: 9)
template mat4(s: string): Expr = Expr(str: s, btype: B_float, vlen: 16)
template disconnected(): Expr = Expr(str: "error", btype: B_float, vlen: 0)
template is_disconnected(e: Expr): bool = e.vlen == 0

proc flt(e: Expr): Expr =
    result = e
    result.vlen = 1
    # use dot product to average vector elements or convert color to bw
    # TODO: cache it !!
    if e.vlen == 4: result.str = &"dot({e.str}.rgb, vec3(0.2126, 0.7152, 0.0722))"
    elif e.vlen > 1: result.str = &"dot({e.str}, vec{e.vlen}({1/3}))"
    # TODO: option to just use single element instead of converting?
    # TODO: output float directly from single channel textures

proc vec2(e: Expr): Expr =
    result = e
    result.vlen = 2
    if e.vlen < 2: result.str = fmt"vec2({e.str})"
    elif e.vlen > 2: result.str = e.str & ".xy"

proc vec3(e: Expr): Expr =
    result = e
    result.vlen = 3
    if e.vlen == 1: result.str = fmt"vec3({e.str})"
    elif e.vlen == 2: result.str = fmt"vec3({e.str},0.0)"
    elif e.vlen > 3: result.str = e.str & ".xyz"

proc vec4(e: Expr): Expr =
    result = e
    result.vlen = 4
    if e.vlen == 1: result.str = fmt"vec4(vec3({e.str}),1.0)"
    elif e.vlen == 2: result.str = fmt"vec4({e.str},0.0,1.0)"
    elif e.vlen == 3: result.str = fmt"vec4({e.str},1.0)"

proc vec3_or_4(e: Expr): Expr =
    result = e
    result.vlen = 3
    if e.vlen == 1: result.str = fmt"vec3({e.str})"
    elif e.vlen == 2: result.str = fmt"vec3({e.str},0.0)"


const GLTYPES = ["","float","vec2","vec3","vec4",
    "","","","","mat3","","","","","","","mat4"]

const GLTYPES_NO_F = ["","","vec2","vec3","vec4",
    "","","","","mat3","","","","","","","mat4"]

type InputMapper = object
    node: FNode
    first: FNode
    first_output: FNode

type NodeBuildContext = object
    tmp_vars: int
    cache: Table[string, seq[Expr]]
    cache_nodes: Table[FNode, seq[Expr]]
    code: seq[string]
    indent: int
    textures: OrderedTable[string, string]
    texture_pixels: OrderedTable[string, TexturePixels]
    uniforms: OrderedSet[Expr]
    varyings: OrderedTable[(VaryingType, Expr), Varying]
    libraries: OrderedSet[string]
    used_functions: OrderedSet[string]
    is_background: bool
    tone_mapping_function: string
    current_group: InputMapper
    group_stack: seq[InputMapper]

var context* {.threadvar.}: NodeBuildContext

proc add_code(s: string) =
    for line in s.split("\n"):
        context.code.add " ".repeat(context.indent) & line

proc add_functions(func_names: varargs[string]) =
    for name in func_names:
        context.used_functions.incl name

proc clear_context() =
    context = typeof(context).default

proc temp(prefix: string): string =
    result = prefix & "_" & $context.tmp_vars
    context.tmp_vars += 1

proc flttmp(prefix="flt"): Expr =
    result = flt(temp(prefix))
    add_code "float " & result.str & ";"

proc vec2tmp(prefix="v2"): Expr =
    result = vec2(temp(prefix))
    add_code "vec2 " & result.str & ";"

proc vec3tmp(prefix="v3"): Expr =
    result = vec3(temp(prefix))
    add_code "vec3 " & result.str & ";"

proc vec4tmp(prefix="v4"): Expr =
    result = vec4(temp(prefix))
    add_code "vec4 " & result.str & ";"

proc tmp_of_type(e: Expr, prefix="tmp"): Expr =
    result = e
    result.str = temp(prefix)
    add_code GLTYPES[result.vlen] & " " & result.str & ";"

var node_functions: Table[string, proc(ins: InputMapper): seq[Expr]]

template toInputMapper(node: FNode): InputMapper =
    InputMapper(node: node, first: node.inputs.first, first_output: node.outputs.first)

template toOutputMapper(node: FNode): InputMapper =
    InputMapper(node: node, first: node.outputs.first)

template outputs(mapper: InputMapper): InputMapper =
    InputMapper(node: mapper.node, first: mapper.node.outputs.first)

iterator items(self: InputMapper): string =
    # Repeating names should append a number at the end
    # It checks the last two names so you can have a repeating A, B, A, B...
    # Note: This should match the logic in `[]` below!
    var inp = self.first
    var last1, last2 = ""
    var i1, i2 = 0
    while inp.valid:
        let name = inp.name.str
        if name == last1:
            i1 += 1
            yield name & $i1
        elif name == last2:
            i2 += 1
            yield name & $i2
        else:
            i2 = 0
            yield name
        swap(i1, i2)
        swap(last1, last2)
        last2 = name
        inp = inp.next

proc getOutputs(node: FNode): seq[Expr]

proc getMutedOutputs(node: Fnode): seq[Expr]

proc getLinkExpr(link: FNode): Expr =
    let fromnode = link.fromnode
    let fromsock = link.fromsock
    var index = 0
    var output = fromnode.outputs.first
    while output.valid and output != fromsock:
        index += 1
        output = output.next
    assert output.valid
    let outputs = getOutputs(fromnode)
    # TODO: this should never happen
    if index > outputs.high:
        return vec3("vec3(1,0,1)")
    return outputs[index]

proc `[]`(self: InputMapper, name: string, vec_len: uint32 = 0): Expr =
    # Note: Name disambiguation should match the logic in items() above!
    var inp = self.first
    var last1, last2 = ""
    var i1, i2 = 0
    while inp.valid:
        let iname = inp.name.str
        let i = if iname == last1:
            i1 += 1
            i1
        elif iname == last2:
            i2 += 1
            i2
        else:
            i2 = 0
            0
        swap(i1, i2)
        swap(last1, last2)
        last2 = iname
        if iname == name or iname & $i == name:
            let default_value = inp.default_value
            let (btype, vlen, vkey) = if default_value.valid:
                let vkey = if "value" in default_value: "value"
                else: "value_euler"
                let value = default_value[vkey]
                (basic_type(value), value.len, vkey)
            else:
                # Most likely a shader input
                (B_float, 4'u32, "")
            if vec_len != 0 and vlen != vec_len:
                inp = inp.next
                continue
            var link = inp.link
            if link.valid:
                let e = link.getLinkExpr
                if not e.is_disconnected:
                    return e
            return if default_value.valid:
                case btype:
                of B_float:
                    Expr(str: fmt"""{GLTYPES_NO_F[vlen]}({
                        ($default_value[vkey].f32[0 ..< vlen.int])[2 .. ^2]
                    })""", btype: B_float, vlen: vlen)
                else:
                    echo "MISSING! " & $btype
                    vec4("missing " & $btype)
            else:
                # Most likely a shader input (TODO: is alpha 0 or 1?)
                vec4("vec4(0)")
        inp = inp.next
    echo "Available keys:"
    for k in self:
        echo "  ",k
    raise newException(KeyError, fmt"key {name} not found")

template `.`*(self: InputMapper, field: untyped): Expr = self[astToStr(field)]

# see mkOutput() below
proc mkOutputImpl(code: string, vars: varargs[Expr]): seq[Expr] =
    var code = code
    if vars.high >= 10:
        for i,e in vars[10 .. vars.high]:
            code = code.replace("$" & $i, e.str)
    for i,e in vars[0 .. min(vars.high, 9)]:
        code = code.replace("$" & $i, e.str)
    if code != "":
        add_code code
    for v in vars:
        result.add v

# identical code but with seq
proc mkOutputImpl(code: string, vars: seq[Expr]): seq[Expr] =
    var code = code
    if vars.high >= 10:
        for i,e in vars[10 .. vars.high]:
            code = code.replace("$" & $i, e.str)
    for i,e in vars[0 .. min(vars.high, 9)]:
        code = code.replace("$" & $i, e.str)
    if code != "":
        add_code code
    for v in vars:
        result.add v

# this template allows us to not create temporary output variables
# when the code is already in cache
template mkOutput(code: string, vars: varargs[Expr]): seq[Expr] =
    if code != "" and code in context.cache:
        context.cache[code]
    else:
        let ovars = mkOutputImpl(code, vars)
        context.cache[code] = ovars
        ovars

# identical code but with seq
template mkOutput(code: string, vars: seq[Expr]): seq[Expr] =
    if code != "" and code in context.cache:
        context.cache[code]
    else:
        let ovars = mkOutputImpl(code, vars)
        context.cache[code] = ovars
        ovars


proc texture_uniform(name: string, gl_type = ""): string =
    if name in context.textures:
        return context.textures[name]
    result = "samp" & $context.textures.len
    context.textures[name] = result

proc texture_uniform(texture_pixels: TexturePixels): string =
    let name = &"ramp_{texture_pixels.hash.int.toHex}"
    context.texture_pixels[name] = texture_pixels
    return texture_uniform(name)

# usage: uniform(vec3("foo"))
proc uniform(e: Expr): Expr =
    context.uniforms.incl e
    return e

proc varying(vtype: VaryingType, att_expr: Expr): Expr =
    let t = (vtype, att_expr)
    let v = if t notin context.varyings:
        let varname = "var" & $context.varyings.len
        let v = Varying(vtype: vtype, varname: varname, attname: att_expr.str,
        gltype: GLTYPES[att_expr.vlen], multiplier: 1)
        context.varyings[t] = v
        v
    else:
        context.varyings[t]
    result = att_expr
    result.str = v.varname

proc getOutputs(node: FNode): seq[Expr] =
    # TODO: table literal for cstrings instead of strings
    let id = node.idname.str
    assert id in node_functions, "node not found: " & id
    let flags = node.flag.i32[0]
    if not flags.testBit(9):
        if node notin context.cache_nodes:
            context.cache_nodes[node] = node_functions[id](node.toInputMapper)
        return context.cache_nodes[node]
    else:
        return node.getMutedOutputs
    
    # TODO: system for knowing which outputs we actually need
    # (for when we can separate a node)
    # if we return all outputs we ignored the value
    # if we return one output we used it

proc getMutedOutputs(node: Fnode): seq[Expr] =
    var first_connected: FNode
    var outp = node.outputs.first
    let is_tex_node = node.idname.str.startswith "ShaderNodeTex"
    while outp.valid:
        let otype = outp.idname.cstr
        if otype == "NodeSocketShader" or is_tex_node:
            result.add flt("0")
            outp = outp.next
            continue
        var inp = node.inputs.first
        var connected: FNode
        while inp.valid:
            if inp.link.valid and inp.name.cstr != "Factor":
                if not first_connected.valid:
                    first_connected = inp
                if inp.idname.cstr == otype:
                    connected = inp
                    break
            inp = inp.next
        if not connected.valid:
            connected = first_connected
        if connected.valid:
            result.add connected.link.getLinkExpr
        else:
            result.add flt("0")
        outp = outp.next

when defined(GEN_MISSING_NODES):
    proc generate_missing_nodes_code(nodes: FNode) =
        echo "MISSING NODES?"
        var node = nodes.first
        while node.valid:
            let id = node.idname.str
            let ins = node.toInputMapper
            let outs = node.toOutputMapper
            if id notin node_functions:
                node_functions[id] = proc(ins: InputMapper): seq[Expr] {.closure.} = @[]
                echo &"\"{id}\": proc(ins: InputMapper): seq[Expr] {{.closure.}} ="
                for name in ins:
                    let i = ins[name]
                    case i.btype:
                    of B_float:
                        echo &"    let {name.toLower} = ins[\"{name}\"] # size {i.vlen}"
                    else:
                        echo "    TODO ", (name, i.btype, i.vlen)
                var out_vars: seq[string]
                for name in outs:
                    let o = outs[name]
                    case o.btype:
                    of B_float:
                        out_vars.add ["","flttmp()","vec2tmp()","vec3tmp()","vec4tmp()"][o.vlen]
                    else:
                        echo "    TODO ", (name, o.btype, o.vlen)
                let tmp_vars = out_vars.join ", "
                echo &"    return mkOutput(&\"\", {tmp_vars})"
                echo &","
            node = node.next
        echo "END MISSING NODES"

type MakeMaterialSettings* = object
    can_be_double_sided*: bool
    tone_mapping_function*: string
    uniform_blocks*: string
    textures_sampler_type*: Table[string, string]
    get_library*: proc(name: string): seq[string]
    get_node_tree*: proc(name: string): FNode

var get_node_tree*: proc(name: string): FNode

proc makeMaterial*(self: BlendLoader;
        bmat: FNode,
        shader_library=self.shader_library,
        settings = MakeMaterialSettings(),
    ): (string, seq[Varying], OrderedTable[string, string], OrderedTable[string, TexturePixels]) =

    get_node_tree = settings.get_node_tree

    # TODO: delete old cached materials?
    let cache_key = (bmat, shader_library, settings.can_be_double_sided)
    if cache_key in self.cached_materials:
        return self.cached_materials[cache_key]

    when defined(GEN_MISSING_NODES):
        generate_missing_nodes_code(bmat.nodetree.nodes)

    clear_context()
    context.tone_mapping_function = settings.tone_mapping_function
    # find output node
    var output_node: FNode
    var node = bmat.nodetree.nodes.first
    
    while node.valid:
        case node.idname.str:
        of "ShaderNodeOutputMaterial":
            output_node = node
        of "ShaderNodeOutputWorld":
            output_node = node
            context.is_background = true
        of "NodeGroupOutput":
            output_node = node
        node = node.next
    if output_node.valid:
        context.indent += 4
        discard output_node.get_outputs()
        context.indent -= 4
        var code = @[
            "precision highp float;",
            if settings.can_be_double_sided:
                "#define IS_FRONT_FACE gl_FrontFacing"
            else:
                "#define IS_FRONT_FACE true"
        ]
        if settings.uniform_blocks.len != 0:
            code.add settings.uniform_blocks
        var uniforms: seq[string]
        var varyings: seq[Varying] = @[
            # Varying(vtype: WorldPosition, varname: "varpos"),
            # Varying(vtype: WorldNormal, varname: "varnormal"),
        ]
        var ext_added = false
        for k,u in context.textures:
            let gl_type = if k in settings.textures_sampler_type:
                settings.textures_sampler_type[k]
            else: "sampler2D"
            if gl_type == "samplerExternalOES":
                ext_added = true
                code.insert("#extension GL_OES_EGL_image_external_essl3 : require", 0)
            code.add fmt"uniform {gl_type} {u};"
        for u in context.uniforms:
            uniforms.add u.str
            # code.add fmt"uniform {GLTYPES[u.vlen]} {u};"
        when not defined(release):
            let block_words = settings.uniform_blocks.split {' ', '\t', ',', ';'}
            for u in context.uniforms:
                assert u.str in block_words, &"Uniform {u.str} not in render blocks"
        for _,v in context.varyings:
            varyings.add v
            code.add fmt"in {v.gltype} {v.varname};"
        # code.add @[
        #     "in vec3 varpos, varnormal;",
        # ]
        for fun in context.used_functions:
            if glsl_functions.glsl_functions.hasKey fun:
                code.add glsl_functions.glsl_functions[fun]
        code.add shader_library
        for name in context.libraries:
            code.add settings.get_library(name)
        code.add @[
            "out vec4 outColor;",
            "void main(){",
        ]
        code.add context.code
        code.add "}"
        let fragment = code.join("\n")
        result = (fragment, varyings, context.textures, context.texture_pixels)
        self.cached_materials[cache_key] = result
    
    clear_context()

# cache-able internal functions

proc world2view_dir(e: Expr): Expr =
    discard uniform(mat4("view_matrix"))
    return mkOutput(&"$0 = (view_matrix * vec4({e}, 0.0)).xyz;", vec3tmp("wo2vi"))[0]

proc view2world_dir(e: Expr): Expr =
    # let to_world = uniform(mat3("view_matrix_inverse3"))
    # return mkOutput(&"$0 = {to_world} * {e};", vec3tmp("vi2wo"))[0]
    discard uniform(mat4("view_matrix_inverse"))
    return mkOutput(&"$0 = (view_matrix_inverse * vec4({e}, 0.0)).xyz;", vec3tmp("wo2vi"))[0]

proc normalize(e: Expr): Expr =
    return mkOutput(&"$0 = normalize({e});", vec3tmp("norm"))[0]

proc invert(e: Expr): Expr =
    return mkOutput(&"$0 = inverse({e});", vec3tmp("inv"))[0]

proc clamp(e: Expr, a,b: float): Expr =
    return mkOutput(&"$0 = clamp({e}, {a:0.7f}, {b:0.7f});", tmp_of_type(e, "clamp"))[0]

proc mix(a,b,t: Expr): Expr =
    return mkOutput(&"$0 = mix({a}, {b}, {t});", tmp_of_type(a, "mix"))[0]

proc min(a,b: Expr): Expr =
    return mkOutput(&"$0 = min({a}, {b});", tmp_of_type(a, "min"))[0]

proc max(a,b: Expr): Expr =
    return mkOutput(&"$0 = max({a}, {b});", tmp_of_type(a, "max"))[0]

proc reflect(a, b: Expr): Expr =
    return mkOutput(&"$0 = reflect({a}, {b});", vec3tmp("refl"))[0]

proc euler_xyz_to_mat3(e: Expr): Expr =
    # TODO: precompute constants? or always assume glsl compiler will do it?
    add_functions "euler_xyz_to_mat3"
    return mkOutput(&"$0 = euler_xyz_to_mat3({e});", tmp_of_type(mat3(""), "em3"))[0]

proc color_to_vectors(e: Expr): Expr =
    return mkOutput(&"$0 = 2.0*{e.vec3} - vec3(1);", vec3tmp("nmap"))[0]

proc view_position(): Expr =
    return varying(ViewPosition, vec3("view_position"))

proc world_position(): Expr =
    return varying(WorldPosition, vec3("world_position"))

proc view_normal(): Expr =
    let n = varying(ViewNormal, vec3("view_normal"))
    return mkOutput(&"$0 = IS_FRONT_FACE?{n}:-{n};", vec3tmp("vnorm"))[0]

proc world_normal(): Expr =
    let n = varying(WorldNormal, vec3("world_normal"))
    return mkOutput(&"$0 = IS_FRONT_FACE?{n}:-{n};", vec3tmp("wnorm"))[0]

proc incoming(): Expr =
    let mat = uniform(mat4("projection_matrix"))
    return mkOutput(&"$0 = ({mat}[3][3] == 0.0) ? {view_position().normalize} : vec3(0.0, 0.0, -1.0);", vec3tmp("inc"))[0]

proc world_incoming(): Expr =
    return vec3(&"-{incoming().view2world_dir}")

proc window_coord(): Expr =
    return mkOutput(&"$0 = vec3(gl_FragCoord.xy * viewport_size_inv, 0.0);", vec3tmp("window"))[0]

# helpers for default socket values when disconnected

proc normal_socket_to_view(normal: Expr): Expr =
    if $normal == "vec3(0.0, 0.0, 0.0)":
        # it is disconnected
        return view_normal().normalize()
    # normal is in world space, we need it in view space
    return normal.vec3.world2view_dir

proc normal_socket_to_world(normal: Expr, normalize: bool=false): Expr =
    if $normal == "vec3(0.0, 0.0, 0.0)":
        # it is disconnected
        return world_normal().normalize()
    if normalize:
        return normal.vec3.normalize()
    return normal.vec3

proc uv_socket(uv: Expr): Expr =
    if $uv == "vec3(0.0, 0.0, 0.0)":
        # it is disconnected
        return varying(Uv, vec2("uv_active"))
    return uv.vec2

proc is_vec3_socket_disconnected(e: Expr): bool =
    $e == "vec3(0.0, 0.0, 0.0)"

node_functions = {

    "ShaderNodeOutputWorld": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let surface = ins["Surface"]
        # let volume = ins["Volume"] # vec4
        return mkOutput(&"outColor = {context.tone_mapping_function}({ins.Surface.vec4});")
    ,

    "NodeReroute": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return @[ins["Input"]]
    ,
    "ShaderNodeOutputMaterial": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return mkOutput(&"outColor = {context.tone_mapping_function}({ins.Surface.vec4});")
    ,
    "ShaderNodeTexImage": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO: alpha output?
        let img = ins.node.id
        if not img.valid:
            return @[vec4("vec4(0)")]
        let s = texture_uniform(img.name.str.strip2)
        var v = ins["Vector"].uv_socket
        # TODO: output float directly from single channel textures
        return mkOutput(&"$0 = texture({s}, {vec2(v)});", vec4tmp())
    ,
    "ShaderNodeTexEnvironment": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let img = ins.node.id
        if not img.valid:
            return @[vec4("vec4(0)")]
        let samp = texture_uniform(img.name.str.strip2)
        let storage = ins.node.storage
        var v = ins["Vector"].vec3
        if v.is_vec3_socket_disconnected:
            if context.is_background:
                # TODO: would it be better to use world_position - camera_world_position?
                v = view_position().view2world_dir()
            else:
                v = world_position()
        v = v.normalize()
        # 0 Linear, 1 Closest, 2 Cubic, 3 Smart
        let interpolation = storage.interpolation.i32[0]
        # 0 equirectangular, 1 mirror ball
        let projection = storage.projection.i32[0]

        if projection == 0:
            # TODO: calculate texture lod
            return mkOutput(dedent &"""
                $0 = textureLod({samp}, vec2(
                    -atan({v}.y, {v}.x) / (2.0 * {PI}) + 0.5,
                    atan({v}.z, length({v}.xy)) / {PI} + 0.5
                ), 0.0);
            """, vec4tmp())
    ,

    #----- INPUT NODES -----#
    
    "ShaderNodeVertexColor": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let name = ins.node.id.name.str.strip2
        let attr = if name != "": "vc_" & name else: "vc_active"
        let v = varying(VertexColor, vec4(attr))
        # TODO: is the color output a vec3 or a vec4??
        return mkOutput(&"$0 = {v}.rgb; $1 = {v}.a;", vec3tmp(), flttmp())
    ,

    "ShaderNodeRGB": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return @[ins.outputs["Color"]]
    ,

    "ShaderNodeValue": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return @[ins.outputs["Value"]]
    ,
    "ShaderNodeAmbientOcclusion": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"]
        let distance = ins["Distance"]
        let normal = ins["Normal"]
        # TODO
        return mkOutput(&"", vec4("vec4(1)"), flt("1.0"))
    ,
    "ShaderNodeAttribute": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let storage = ins.node.storage
        let name = storage.name.str.strip2
        let attr = if name != "": "vc_" & name else: "vc_active"
        let v = varying(VertexColor, vec4(attr))
        # TODO: other types of attribute
        return mkOutput(&"", v, v.vec3, v.vec2, v.flt)
    ,
    "ShaderNodeBevel": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let radius = ins["Radius"]
        var normal = ins["Normal"].normal_socket_to_world
        return @[normal.vec3]
    ,
    "ShaderNodeCameraData": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO: separate in 3 mkoutputs, see if we can remove them if unused
        let pos = view_position()
        return mkOutput(dedent &"""
            $0 = {pos.normalize}*vec3(1.0,1.0,-1.0);
            $1 = length({pos});
            $2 = abs({pos}.z);
        """, vec3tmp(), flttmp(), flttmp())
    ,
    "ShaderNodeHairInfo": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return mkOutput(&"", flt("0.0"), flt("0.0"), flt("0.0"), flt("0.0"), vec3("vec3(0.0)"), flt("0.0"))
    ,
    "ShaderNodeFresnel": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let ior = ins["IOR"].flt
        var normal = ins["Normal"].normal_socket_to_view.normalize
        add_functions "fresnel_dielectric"
        return mkOutput(&"$0 = fresnel_dielectric({incoming()}, {normal}, IS_FRONT_FACE? {ior} : 1.0/{ior});", flttmp())
    ,
    "ShaderNodeNewGeometry": proc(ins: InputMapper): seq[Expr] {.closure.} =
        if context.is_background:
            let wp = view_position().view2world_dir.normalize
            let wpn = vec3("-" & $wp)
            return @[
                wp,
                wpn,
                vec3("vec3(0)"),
                wpn,
                wpn,
                vec3("vec3(0)"),
                flt("(0.0)"), flt("(0.0)"), flt("(0.0)"), 
            ]
        return mkOutput(dedent &"""
            $0 = {world_position()};
            $1 = {world_normal().normalize()};
            $3 = {world_normal().normalize()};
            $4 = {world_incoming()};
            $6 = IS_FRONT_FACE? 0.0 : 1.0;
        """, vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), flttmp(), flttmp(), flttmp())
    ,
    "ShaderNodeLayerWeight": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let blend = ins["Blend"].flt
        var normal = ins["Normal"].normal_socket_to_view.normalize
        let eta = mkOutput(&"$0 = max(1.0 - {blend}, 0.00001);", flttmp())[0]
        let tmp = flttmp()
        add_functions "fresnel_dielectric"
        return mkOutput(dedent &"""
            $0 = fresnel_dielectric({incoming()}, {normal}, IS_FRONT_FACE? 1.0/{eta} : {eta});
            $1 = abs(dot({incoming()}, {normal}));
            if ({blend} != 0.5) {{
                {tmp} = clamp({blend}, 0.0, 0.99999);
                {tmp} = ({tmp} < 0.5) ? 2.0 * {tmp} : 0.5 / (1.0 - {tmp});
                $1 = pow($1, {tmp});
            }}
            $1 = 1.0 - $1;
        """, flttmp(), flttmp())
    ,
    "ShaderNodeObjectInfo": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO: random
        return mkOutput(dedent &"""
            $0 = {uniform(mat4("model_matrix"))}[3].xyz;
            $1 = {uniform(vec4("object_color"))};
            $2 = {uniform(vec4("object_color"))}.a;
        """, vec3tmp(), vec4tmp(), flttmp(), flttmp(), flttmp(), flttmp())
    ,
    "ShaderNodeParticleInfo": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO
        return mkOutput(&"", flttmp(), flttmp(), flttmp(), flttmp(), vec3tmp(), flttmp(), vec3tmp(), vec3tmp())
    ,
    "ShaderNodePointInfo": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO
        return mkOutput(&"", vec3tmp(), flttmp(), flttmp())
    ,
    "ShaderNodeTangent": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let props = ins.node.storage
        case props.direction_type.i32[0]:
            of 0: # radial
                var orco = varying(Orco, vec3("orco"))
                orco = case props.axis.i32[0]:
                    of 0: mkOutput(&"$0 = vec3(0.0, {orco}.z * -0.5, {orco}.y * 0.5);", vec3tmp())[0]
                    of 1: mkOutput(&"$0 = vec3({orco}.z * -0.5, 0.0, {orco}.x * 0.5);", vec3tmp())[0]
                    else: mkOutput(&"$0 = vec3({orco}.y * -0.5, {orco}.x * 0.5, 0.0);", vec3tmp())[0]
                return mkOutput(dedent &"""
                    $0 = ({uniform(mat4("model_matrix"))}*vec4({orco}, 0.0)).xyz;
                    $0 = normalize(cross({world_normal()}, cross($0, {world_normal()})));
                """, vec3tmp())
            of 1: # uv
                let uv_map = props.uv_map.str
                let attr_tg = if uv_map != "":
                    varying(Tangent, vec4("tg_" & uv_map.replace(re"[^_A-Za-z0-9]", "")))
                else:
                    varying(Tangent, vec4("tg_active"))
                return @[attr_tg.vec3.view2world_dir]
            else:
                raise newException(ValueError, fmt"Unknown tangent type {props.direction_type.i32[0]}")
    ,
    "ShaderNodeTexCoord": proc(ins: InputMapper): seq[Expr] {.closure.} =
        if context.is_background:
            let wp = view_position().view2world_dir.normalize
            let wpn = vec3("-" & $wp)
            let vp = $view_position().normalize
            let vpn = vec3(&"vec3({vp}.x, {vp}.y, -{vp}.z)")
            return @[
                wp, wpn, vec3("vec3(0)"), wp, vpn, 
                window_coord(), wpn,
            ]
        # TODO: object coordinates, window
        let props = ins.node.storage
        let vpos = view_position()
        let uv_attr = varying(Uv, vec2("uv_active"))
        let orco = varying(Orco, vec3("orco"))
        let ob_normal = varying(ObjectNormal, vec3("object_normal"))
        return mkOutput(dedent &"""
            $0 = {orco} * 0.5 + vec3(0.5);
            $1 = IS_FRONT_FACE ? {ob_normal}:-{ob_normal};
            $2 = vec3({uv_attr}, 0);
            $4 = vec3({vpos}.xy, -{vpos}.z);
            $5 = {window_coord()};
            $6 = {reflect(incoming(), view_normal().normalize).view2world_dir};
            // $7 = {{uniform(mat4("projection_matrix"))}} * vec4({vpos}, 1.0);
        """, vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(), vec3tmp(),
            # $7{color}
            vec4tmp())[0 .. 6]
    ,
    "ShaderNodeUVMap": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let props = ins.node.storage
        let uv_map = props.uv_map.str
        return if uv_map != "":
            @[varying(Uv, vec2("uv_" & uv_map.replace(re"[^_A-Za-z0-9]", "")))]
        else:
            @[varying(Uv, vec2("uv_active"))]
    ,
    "ShaderNodeVolumeInfo": proc(ins: InputMapper): seq[Expr] {.closure.} =
        return mkOutput(&"", vec4tmp(), flttmp(), flttmp(), flttmp())
    ,
    "ShaderNodeWireframe": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let size = ins["Size"]
        return mkOutput(&"", flttmp())
    ,

    #----- COLOR NODES -----#

    "ShaderNodeBrightContrast": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec4
        let brightness = ins["Bright"].flt
        let contrast = ins["Contrast"].flt
        return mkOutput(dedent &"""
            $1 = 1.0 + {contrast};
            $2 = {brightness} - {contrast} * 0.5;

            $0.r = max($1 * {color}.r + $2, 0.0);
            $0.g = max($1 * {color}.g + $2, 0.0);
            $0.b = max($1 * {color}.b + $2, 0.0);
            $0.a = {color}.a;
        """, vec4tmp(),
        flttmp(), flttmp())[0 .. 0]
    ,
    "ShaderNodeBsdfDiffuse": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let base = ins["Color"].vec3
        let normal = ins["Normal"].normal_socket_to_world
        
        add_functions "diffuse_node"
        return mkOutput(&"$0 = diffuse_node({base}, " &
            &"{world_incoming()}, " &
            &"{normal}, " &
            &"{world_position()} " &
            &");", vec4tmp())
    ,
    "ShaderNodeGamma": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # NOTE: ignoring compatibility with gamma <= 0
        let color = ins["Color"].vec4
        let gamma = ins["Gamma"].flt
        return mkOutput(dedent &"""
            $0 = {color};
            if ({color}.r > 0.0) $0.r = pow({color}.r, {gamma});
            if ({color}.g > 0.0) $0.g = pow({color}.g, {gamma});
            if ({color}.b > 0.0) $0.b = pow({color}.b, {gamma});
        """, vec4tmp())
    ,
    "ShaderNodeHueSaturation": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let hue = ins["Hue"]
        let sat = ins["Saturation"]
        let value = ins["Value"]
        let fac = ins["Fac"]
        let color = ins["Color"].vec4
        add_functions("rgb_to_hsv", "hsv_to_rgb")
        return mkOutput(dedent &"""
            $0 = rgb_to_hsv({color});
            $0[0] = fract($0[0] + {hue} + 0.5);
            $0[1] = clamp($0[1] * {sat}, 0.0, 1.0);
            $0[2] = $0[2] * {value};
            $0 = hsv_to_rgb($0);
            $0 = mix({color}, $0, {fac});
        """, vec4tmp())
    ,
    "ShaderNodeInvert": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let fac = ins["Fac"].flt
        let color = ins["Color"].vec4
        return mkOutput(&"$0 = vec4(mix({color}.rgb, vec3(1)-{color}.rgb, {fac}), {color}.a);", vec4tmp())
    ,
    "ShaderNodeLightFalloff": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO: This is for lamps
        let strength = ins["Strength"]
        let smooth = ins["Smooth"]
        return mkOutput(&"", flttmp(), flttmp(), flttmp())
    ,
    "ShaderNodeMix": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let props = ins.node.storage
        let blend_type = props.blend_type.i8[0]
        let clamp_factor = props.clamp_factor.i8[0]
        let clamp_result = props.clamp_result.i8[0]
        let data_type = props.data_type.i8[0]
        let factor_mode = props.factor_mode.i8[0]
        case data_type:
            of 0: # float
                var fac = ins["Factor"].flt
                if clamp_factor == 1:
                    fac = clamp(fac, 0, 1)
                let a = ins["A"].flt
                let b = ins["B"].flt
                return mkOutput(&"$0 = mix({a}, {b}, {fac});", flttmp("a"), vec3tmp("b"), vec4tmp("c"), vec3tmp("d"))
            of 1: # vector
                # TODO: IS THIS WRONG? It's slightly different with e.g. normal or tangent
                var fac = if factor_mode == 0: # uniform
                    ins["Factor"].flt
                else: # non-uniform
                    ins["Factor1"].vec3
                if clamp_factor == 1:
                    fac = clamp(fac, 0, 1)
                let a = ins["A1"].vec3
                let b = ins["B1"].vec3
                return mkOutput(&"$1 = mix({a}, {b}, {fac});", flttmp("e"), vec3tmp("f"), vec4tmp("g"), vec3tmp("h"))
            of 2: # rgba
                var fac = ins["Factor"].flt
                if clamp_factor == 1:
                    fac = clamp(fac, 0, 1)
                var clamp_l, clamp_r = ""
                if clamp_result == 1:
                    clamp_l = "clamp("
                    clamp_r = ", 0.0, 1.0)"
                let a4 = ins["A2"].vec4
                let b4 = ins["B2"].vec4
                let a = a4.vec3
                let b = b4.vec3
                var has_alpha = false
                let code = case blend_type:
                    of 0: # MIX
                        &"vec4(mix({a}, {b}, {fac}), {a4}.a)"
                    of 1: # ADD
                        &"vec4(mix({a}, {a}+{b}, {fac}), {a4}.a)"
                    of 2: # MULTIPLY
                        &"vec4(mix({a}, {a}*{b}, {fac}), {a4}.a)"
                    of 3: # SUBTRACT
                        &"vec4(mix({a}, {a}-{b}, {fac}), {a4}.a)"
                    of 4: # SCREEN
                        &"vec4(vec3(1) - (vec3(1.0-{fac}) + {fac} * (vec3(1) - {b})) * (vec3(1) - {a}), {a4}.a)"
                    of 5: # DIVIDE
                        dedent &"""vec4(mix({a}, vec3(
                            ({b}.r != 0.0)? {a}.r / {b}.r : {a}.r,
                            ({b}.g != 0.0)? {a}.g / {b}.g : {a}.g,
                            ({b}.b != 0.0)? {a}.b / {b}.b : {a}.b
                        ), {fac}), {a4}.a)"""
                    of 6: # DIFFERENCE
                        &"vec4(mix({a}, abs({a}-{b}), {fac}), {a4}.a)"
                    of 7: # DARKEN
                        &"vec4(mix({a}, min({a}, {b}), {fac}), {a4}.a)"
                    of 8: # LIGHTEN
                        &"vec4(mix({a}, max({a}, {b}), {fac}), {a4}.a)"
                    of 9: # OVERLAY
                        add_functions("mix_overlay")
                        &"mix_overlay({a4}, {b4}, {fac}, $2)"
                    of 10: # DODGE
                        add_functions("mix_dodge")
                        &"mix_dodge({a4}, {b4}, {fac}, $2)"
                    of 11: # BURN
                        add_functions("mix_burn")
                        &"mix_burn({a4}, {b4}, {fac}, $2)"
                    of 12: # HUE
                        add_functions("rgb_to_hsv", "hsv_to_rgb", "mix_hue")
                        &"mix_hue({a4}, {b4}, {fac}, $2)"
                    of 13: # SATURATION
                        add_functions("rgb_to_hsv", "hsv_to_rgb", "mix_saturation")
                        &"mix_saturation({a4}, {b4}, {fac}, $2)"
                    of 14: # VALUE
                        add_functions("rgb_to_hsv", "hsv_to_rgb", "mix_value")
                        &"mix_value({a4}, {b4}, {fac}, $2)"
                    of 15: # COLOR
                        add_functions("rgb_to_hsv", "hsv_to_rgb", "mix_color")
                        &"mix_color({a4}, {b4}, {fac}, $2)"
                    of 16: # SOFT_LIGHT
                        add_functions("mix_soft_light")
                        &"mix_soft_light({a4}, {b4}, {fac}, $2)"
                    of 17: # LINEAR_LIGHT
                        &"{a4} + {fac} * (2.0 * ({b4} - vec4(0.5)))"
                    else:
                        raise newException(ValueError, fmt"Unknown mix blend type {blend_type}")
                return mkOutput(&"$2 = {clamp_l}{code}{clamp_r};", flttmp("i"), vec3tmp("j"), vec4tmp("k"), vec3tmp("l"))
            else:
                raise newException(ValueError, fmt"Unknown mix data type {data_type}")
        
    ,
    "ShaderNodeBsdfPrincipled": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let base = ins["Base Color"].vec3
        let metallic = ins["Metallic"].flt
        let roughness = ins["Roughness"].flt
        let ior = ins["IOR"].flt
        # let alpha = ins["Alpha"].flt
        let normal = ins["Normal"].normal_socket_to_world
        # let weight1 = ins["Weight"].flt
        # let subsurface_weight = ins["Subsurface Weight"].flt
        # let subsurface_radius = ins["Subsurface Radius"].vec3
        # let subsurface_scale = ins["Subsurface Scale"].flt
        # let subsurface_ior = ins["Subsurface IOR"].flt
        # let subsurface_anisotropy = ins["Subsurface Anisotropy"].flt
        # let specular_ior = ins["Specular IOR Level"].flt
        # let specular_tint = ins["Specular Tint"].vec3
        # let anisotropic = ins["Anisotropic"].flt
        # let anisotropic_rotation = ins["Anisotropic Rotation"].flt
        # let tangent = ins["Tangent"].vec3
        # let transmission_weight = ins["Transmission Weight"].flt
        # let coat_weight = ins["Coat Weight"].flt
        # let coat_roughness = ins["Coat Roughness"].flt
        # let coat_ior = ins["Coat IOR"].flt
        # let coat_tint = ins["Coat Tint"].vec3
        # let coat_normal = ins["Coat Normal"].vec3
        # let sheen_weight = ins["Sheen Weight"].flt
        # let sheen_roughness = ins["Sheen Roughness"].flt
        # let sheen_tint = ins["Sheen Tint"].vec3
        let emission_color = ins["Emission Color"].vec3
        let emission_strength = ins["Emission Strength"].flt
        
        add_functions "principled_node"
        return mkOutput(&"$0 = principled_node({base}, " &
            &"{world_incoming()}, " &
            &"{normal}, " &
            &"{world_position()}, " &
            &"{metallic}, " &
            &"{roughness}, " &
            &"{ior} " &
            &") + vec4({emission_color} * {emission_strength}, 0.0);", vec4tmp())
    ,
    "ShaderNodeRGBCurve": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let fac = ins["Fac"].flt
        let color = ins["Color"].vec4
        var curves: seq[seq[CurveMapPoint]]
        for cm in ins.node.storage.cm:
            if not cm.curve.valid:
                continue
            var points: seq[CurveMapPoint]
            for p in cm.curve[0 ..< cm.totpoint.i16[0]]:
                let ptype = ((p.flag.i16[0] shr 1) and 3).CurveMapPointHandleType
                points.add CurveMapPoint(vec:[
                    (0.0,0.0),
                    (p.x.f32[0].float,p.y.f32[0].float),
                    (0.0,0.0),
                ], h1: ptype, h2: ptype)
            points = calc_all_handles_curvemap(points)
            curves.add points
        let width = 256
        let pixels = curves_to_mapping_float_pixels(curves, width, 0.0, 1.0, 1.0, 0.0)

        let s = texture_uniform(
            TexturePixels(pixels: pixels, width: width, height:1, format: RGBA_f32, tex_type: Tex2D)
        )

        return mkOutput(dedent &"""
            $0 = vec4(
                texture({s}, vec2({color}.r, 0.5)).a,
                texture({s}, vec2({color}.g, 0.5)).a,
                texture({s}, vec2({color}.b, 0.5)).a,
                {color}.a
            );
            $0.r = texture({s}, vec2($0.r, 0.5)).r;
            $0.g = texture({s}, vec2($0.g, 0.5)).g;
            $0.b = texture({s}, vec2($0.b, 0.5)).b;
            $0 = mix({color}, $0, {fac});
        """, vec4tmp())
    ,

    #----- CONVERTER NODES -----#

    "ShaderNodeBlackbody": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let temp = ins["Temperature"]
        add_functions("blackbody")
        return mkOutput(&"$0 = blackbody(max(0.0,{temp.flt}));", vec4tmp())
    ,

    "ShaderNodeClamp": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let is_range = ins.node.custom1.i16[0] == 1
        let value = ins["Value"].flt
        let a = ins["Min"].flt
        let b = ins["Max"].flt
        if is_range:
            return mkOutput(&"$0 = clamp({value}, min({a},{b}), max({a},{b}));", flttmp())
        else:
            return mkOutput(&"$0 = min(max({value}, {a}), {b});", flttmp())
    ,

    "ShaderNodeValToRGB": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # A.K.A. ColorRamp
        let st = ins.node.storage
        let ipotype = st.ipotype.u8[0].RampInterpolationType
        let ipotype_hue = st.ipotype_hue.u8[0].RampHueInterpolationType
        let color_mode = st.color_mode.u8[0].RampColorMode
        var steps: seq[(array[4, float32], float32)]
        for d in st.data[0 ..< st.tot.i16[0]]:
            let c = d.r.f32
            let pos = d.pos.f32[0]
            steps.add ([c[0],c[1],c[2],c[3]], pos)
        let width = 256
        let pixels = if color_mode == Rgb:
            make_ramp_pixels(steps, width, ipotype)
        else:
            make_hue_ramp_pixels(steps, width, color_mode, ipotype_hue)
        let s = texture_uniform(
            TexturePixels(pixels: pixels, width: width, height:1, format: RGBA_f32, tex_type: Tex2D)
        )
        let fac = ins["Fac"].flt()
        return mkOutput(&"$0 = texture({s}, vec2({fac}, 0.5));", vec4tmp())
    ,

    "ShaderNodeCombineColor": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let red = ins["Red"].flt
        let green = ins["Green"].flt
        let blue = ins["Blue"].flt
        let mode = if ins.node.storage.valid:
            ins.node.storage.mode.i8[0]
        else:
            0'i8
        case mode:
            of 0:
                return mkOutput(&"$0 = vec4({red}, {green}, {blue}, 1.0);", vec4tmp())
            of 1:
                add_functions("hsv_to_rgb")
                return mkOutput(&"$0 = hsv_to_rgb(vec4({red}, {green}, {blue}, 1.0));", vec4tmp())
            of 2:
                raise newException(ValueError, fmt"HSL mode not supported")
            else:
                raise newException(ValueError, fmt"Unknown mode {mode}")
    ,

    "ShaderNodeCombineXYZ": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let x = ins["X"].flt
        let y = ins["Y"].flt
        let z = ins["Z"].flt
        return mkOutput(&"$0 = vec3({x}, {y}, {z});", vec3tmp())
    ,

    "ShaderNodeFloatCurve": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let storage = ins.node.storage
        let flags = storage.flag.i32[0]
        let extrapolate = (flags and 16) != 0
        let fac = ins["Factor"].flt
        let value = ins["Value"].flt
        var curves: seq[seq[CurveMapPoint]]
        var start_x = -0.001
        var end_x = 0.001
        let cm = ins.node.storage.cm[0]
        var points: seq[CurveMapPoint]
        for p in cm.curve[0 ..< cm.totpoint.i16[0]]:
            let ptype = ((p.flag.i16[0] shr 1) and 3).CurveMapPointHandleType
            points.add CurveMapPoint(vec:[
                (0.0,0.0),
                (p.x.f32[0].float,p.y.f32[0].float),
                (0.0,0.0),
            ], h1: ptype, h2: ptype)
        points = calc_all_handles_curvemap(points)
        # TODO: if a curve has 2 points or less
        #       use the curve equation exclusively
        start_x = points[0].vec[1][0]
        end_x = points[^1].vec[1][0]
        curves.add points
        let width = 256
        # adjust the texture so the center of first and last pixels match 
        # with the extents (so they span two less halfs of width)
        let px_size = (end_x-start_x)/(width-1).float32
        let tex_start_x = start_x - px_size/2
        let tex_end_x = end_x + px_size/2
        let pixels = curves_to_mapping_float_pixels(curves, width, tex_start_x, tex_end_x, 1.0, 0.0)

        let s = texture_uniform(
            TexturePixels(pixels: pixels, width: width, height:1, format: RGBA_f32, tex_type: Tex2D)
        )

        let in_scale = 1/(tex_end_x-tex_start_x)
        let in_offset = -tex_start_x * in_scale

        if not extrapolate:
            return mkOutput(dedent &"""
                $0 = texture({s}, vec2({value}*{in_scale}+{in_offset}, 0.5)).r;
                $0 = mix({value}, $0, {fac});
            """, flttmp())
        
        let eq = get_line_equations(curves[0][0], curves[0][^1])
        # m and b are the variables of a linear equation (y = mx + b)
        # m is the slope and b is y-intercept
        let left_m = eq[0][0]
        let left_b = eq[0][1]
        let right_m = eq[1][0]
        let right_b = eq[1][1]
        
        return mkOutput(dedent &"""
            $0 = {value} <= {start_x} ? {value}*{left_m }+{left_b} :
                 {value} >= {end_x} ? {value}*{right_m}+{right_b} :
                    texture({s}, vec2({value}*{in_scale}+{in_offset}, 0.5)).r;
            $0 = mix({value}, $0, {fac});
        """, flttmp())
    ,

    "ShaderNodeMapRange": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let st = ins.node.storage
        var data_type, intp_type: uint8
        var use_clamp: bool
        if st.valid:
            data_type = st.data_type.u8[0]
            intp_type = st.interpolation_type.u8[0]
            use_clamp = st.clamp.u8[0] != 0
        else:
            data_type = 10
            intp_type = ins.node.custom2.i16[0].uint8
            use_clamp = ins.node.custom1.i16[0] != 0
        var value, min0, max0, min1, max1, steps: Expr
        let outn = if data_type == 10: # float
            value = ins["Value"].flt
            min0 = ins["From Min", 1].flt
            max0 = ins["From Max", 1].flt
            min1 = ins["To Min", 1].flt
            max1 = ins["To Max", 1].flt
            steps = ins["Steps", 1].flt
            "$0"
        elif data_type == 48: # vec3
            value = ins["Vector"].vec3
            min0 = ins["From Min", 3].vec3
            max0 = ins["From Max", 3].vec3
            min1 = ins["To Min", 3].vec3
            max1 = ins["To Max", 3].vec3
            steps = ins["Steps", 3].vec3
            "$1"
        else:
            raise newException(ValueError, fmt"Unknown data type {data_type} in map range node")
        var funname = case intp_type:
            of 0: "map_range_linear"
            of 1: "map_range_stepped"
            of 2: "map_range_smooth"
            of 3: "map_range_smoother"
            else:
                raise newException(ValueError, fmt"Unknown interpolation type {data_type} in map range node")
        if use_clamp:
            fun_name &= "_clamp"
        add_functions(fun_name)
        return mkOutput(&"{outn} = {fun_name}({value}, {min0}, {max0}, {min1}, {max1}, {steps});", flttmp(), vec3tmp())
    ,

    "ShaderNodeMath": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let operation = ins.node.custom1.i16[0]
        let use_clamp = ins.node.custom2.i16[0] != 0
        let a = ins["Value"].flt
        let b = ins["Value1"].flt
        let c = ins["Value2"].flt
        var v: seq[Expr]
        case operation:
            of 0: # ADD
                v = mkOutput(&"$0 = {a}+{b};", flttmp())
            of 1: # SUBTRACT
                v = mkOutput(&"$0 = {a}-{b};", flttmp())
            of 2: # MULTIPLY
                v = mkOutput(&"$0 = {a}*{b};", flttmp())
            of 3: # DIVIDE
                v = mkOutput(&"$0 = ({b} != 0.0)? {a}/{b} : 0.0;", flttmp())
            of 4: # SINE
                v = mkOutput(&"$0 = sin({a});", flttmp())
            of 5: # COSINE
                v = mkOutput(&"$0 = cos({a});", flttmp())
            of 6: # TANGENT
                v = mkOutput(&"$0 = tan({a});", flttmp())
            of 7: # ARCSINE
                v = mkOutput(&"$0 = ({a} <= 1.0 && {a} >= -1.0) ? asin({a}) : 0.0;", flttmp())
            of 8: # ARCCOSINE
                v = mkOutput(&"$0 = ({a} <= 1.0 && {a} >= -1.0) ? acos({a}) : 0.0;", flttmp())
            of 9: # ARCTANGENT
                v = mkOutput(&"$0 = atan({a});", flttmp())
            of 10: # POWER
                # TODO: use compatible pow and allow negative exponent?
                    # make it optional?
                v = mkOutput(&"$0 = pow({a}, {b});", flttmp())
            of 11: # LOGARITHM
                v = mkOutput(&"$0 = ({a} > 0.0  && {b} > 0.0)?log2({a}/{b}):0.0;", flttmp())
            of 12: # MINIMUM
                v = mkOutput(&"$0 = min({a}, {b});", flttmp())
            of 13: # MAXIMUM
                v = mkOutput(&"$0 = max({a}, {b});", flttmp())
            of 14: # ROUND
                v = mkOutput(&"$0 = floor({a}+0.5);", flttmp())
            of 15: # LESS_THAN
                v = mkOutput(&"$0 = ({a}<{b})?1.0:0.0;", flttmp())
            of 16: # GREATER_THAN
                v = mkOutput(&"$0 = ({a}>{b})?1.0:0.0;", flttmp())
            of 17: # MODULO (truncated)
                v = mkOutput(&"$0 = sign({a})*mod(abs({a}),abs({b}));", flttmp())
            of 18: # ABSOLUTE
                v = mkOutput(&"$0 = abs({a});", flttmp())
            of 19: # ARCTAN2
                v = mkOutput(&"$0 = atan({a},{b});", flttmp())
            of 20: # FLOOR
                v = mkOutput(&"$0 = floor({a});", flttmp())
            of 21: # CEIL
                v = mkOutput(&"$0 = ceil({a});", flttmp())
            of 22: # FRACTION
                v = mkOutput(&"$0 = fract({a});", flttmp())
            of 23: # SQRT
                v = mkOutput(&"$0 = max(0.0, sqrt({a}));", flttmp())
            of 24: # INV_SQRT
                v = mkOutput(&"$0 = inversesqrt({a});", flttmp())
            of 25: # SIGN
                v = mkOutput(&"$0 = sign({a});", flttmp())
            of 26: # EXPONENT
                v = mkOutput(&"$0 = exp({a});", flttmp())
            of 27: # RADIANS
                v = mkOutput(&"$0 = radians({a});", flttmp())
            of 28: # DEGREES
                v = mkOutput(&"$0 = degrees({a});", flttmp())
            of 29: # SINH
                v = mkOutput(&"$0 = sinh({a});", flttmp())
            of 30: # COSH
                v = mkOutput(&"$0 = cosh({a});", flttmp())
            of 31: # TANH
                v = mkOutput(&"$0 = tanh({a});", flttmp())
            of 32: # TRUNC
                v = mkOutput(&"$0 = trunc({a});", flttmp())
            of 33: # SNAP
                v = mkOutput(&"$0 = floor(({b}>0.0)?{a}/{b}:0.0) * {b};", flttmp())
            of 34: # WRAP
                add_functions("wrap")
                v = mkOutput(&"$0 = wrap({a},{b},{c});", flttmp())
            of 35: # COMPARE
                v = mkOutput(&"$0 = (abs({a} - {b}) <= max({c}, 1e-5)) ? 1.0 : 0.0;", flttmp())
            of 36: # MULTIPLY_ADD
                v = mkOutput(&"$0 = {a} * {b} + {c};", flttmp())
            of 37: # PINGPONG
                v = mkOutput(&"$0 = ({b} != 0.0) ? abs(fract(({a} - {b}) / ({b} * 2.0)) * {b} * 2.0 - {b}) : 0.0;", flttmp())
            of 38: # SMOOTH_MIN
                add_functions("math_smoothmin")
                v = mkOutput(&"$0 = math_smoothmin({a}, {b}, {c});", flttmp())
            of 39: # SMOOTH_MAX
                add_functions("math_smoothmin")
                v = mkOutput(&"$0 = -math_smoothmin(-{a}, -{b}, {c});", flttmp())
            of 40: # MODULO (floored)
                v = mkOutput(&"$0 = ({b} != 0.0) ? mod({a},{b}) : 0.0;", flttmp())
            else:
                raise newException(ValueError, fmt"Unknown operation {operation}")
        if use_clamp:
            v[0] = clamp(v[0], 0, 1)
        return v
    ,

    "ShaderNodeRGBToBW": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec3_or_4
        return mkOutput(&"$0 = {color}.r * 0.2126 + {color}.g * 0.7152 + {color}.b * 0.0722;", flttmp())
    ,

    "ShaderNodeSeparateColor": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let c = ins["Color"].vec3_or_4
        let mode = if ins.node.storage.valid:
            ins.node.storage.mode.i8[0]
        else:
            0'i8
        case mode:
            of 0:
                return mkOutput(&"$0 = {c}.r; $1 = {c}.g; $2 = {c}.b;", flttmp(), flttmp(), flttmp())
            of 1:
                add_functions("rgb_to_hsv")
                let hsv = mkOutput(&"$0 = rgb_to_hsv({c});", vec4tmp())[0]
                return mkOutput(&"$0 = {hsv}.x; $1 = {hsv}.y; $2 = {hsv}.z;", flttmp(), flttmp(), flttmp())
            of 2:
                raise newException(ValueError, fmt"HSL mode not supported")
            else:
                raise newException(ValueError, fmt"Unknown mode {mode}")
    ,

    "ShaderNodeSeparateXYZ": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let v = ins["Vector"].vec3
        return mkOutput(&"$0 = {v}.x; $1 = {v}.y; $2 = {v}.z;", flttmp(), flttmp(), flttmp())
    ,

    "ShaderNodeShaderToRGB": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let shader = ins["Shader"].vec4
        return mkOutput(&"$1 = {shader}.a;", shader, flttmp())
    ,

    "ShaderNodeVectorMath": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let operation = ins.node.custom1.i16[0]
        let use_clamp = ins.node.custom2.i16[0] != 0
        let a = ins["Vector"].vec3
        let b = ins["Vector1"].vec3
        let c = ins["Vector2"].vec3
        let s = ins["Scale"].flt
        var vec_out = disconnected()
        var flt_out = disconnected()
        case operation:
        of 0: # ADD
            vec_out = mkOutput(&"$0 = {a}+{b};", vec3tmp())[0]
        of 1: # SUBTRACT
            vec_out = mkOutput(&"$0 = {a}-{b};", vec3tmp())[0]
        of 2: # MULTIPLY
            vec_out = mkOutput(&"$0 = {a}*{b};", vec3tmp())[0]
        of 3: # DIVIDE
            vec_out = mkOutput(&"$0 = {a}/{b};", vec3tmp())[0]
        of 4: # CROSS_PRODUCT
            vec_out = mkOutput(&"$0 = cross({a},{b});", vec3tmp())[0]
        of 5: # PROJECT
            vec_out = mkOutput(&"$0 = project({a},{b});", vec3tmp())[0]
            add_functions("project")
        of 6: # REFLECT
            vec_out = mkOutput(&"$0 = reflect({a}, normalize({b}));", vec3tmp())[0]
        of 7: # DOT_PRODUCT
            flt_out = mkOutput(&"$0 = dot({a},{b});", flttmp())[0]
        of 8: # DISTANCE
            flt_out = mkOutput(&"$0 = distance({a},{b});", flttmp())[0]
        of 9: # LENGTH
            flt_out = mkOutput(&"$0 = length({a});", flttmp())[0]
        of 10:# SCALE
            vec_out = mkOutput(&"$0 = {a}*{s};", vec3tmp())[0]
        of 11:# NORMALIZE
            vec_out = mkOutput(&"$0 = normalize({a});", vec3tmp())[0]
        of 12:# WRAP
            vec_out = mkOutput(&"$0 = floor({a} / {b}) * {b};", vec3tmp())[0]
        of 13:# FLOOR
            vec_out = mkOutput(&"$0 = floor({a});", vec3tmp())[0]
        of 14:# CEIL
            vec_out = mkOutput(&"$0 = ceil({a});", vec3tmp())[0]
        of 15:# MODULO
            vec_out = mkOutput(&"$0 = modulo({a},{b});", vec3tmp())[0]
            add_functions("modulo")
        of 16:# FRACTION
            vec_out = mkOutput(&"$0 = fract({a});", vec3tmp())[0]
        of 17:# ABSOLUTE
            vec_out = mkOutput(&"$0 = abs({a});", vec3tmp())[0]
        of 18:# MINIMUM
            vec_out = mkOutput(&"$0 = min({a},{b});", vec3tmp())[0]
        of 19:# MAXIMUM
            vec_out = mkOutput(&"$0 = max({a},{b});", vec3tmp())[0]
        of 20:# WRAP
            vec_out = mkOutput(&"$0 = vec3(wrap({a}.x, {b}.x, {c}.x), " &
                &"wrap({a}.y, {b}.y, {c}.y), wrap({a}.z, {b}.z, {c}.z));", vec3tmp())[0]
            add_functions("wrap")
        of 21:# SIN
            vec_out = mkOutput(&"$0 = sin({a});", vec3tmp())[0]
        of 22:# COS
            vec_out = mkOutput(&"$0 = cos({a});", vec3tmp())[0]
        of 23:# TAN
            vec_out = mkOutput(&"$0 = tan({a});", vec3tmp())[0]
        of 24:# REFLECT
            vec_out = mkOutput(&"$0 = refract({a}, normalize({b}), {s});", vec3tmp())[0]
        of 25:# FACEFORWARD
            vec_out = mkOutput(&"$0 = {a} * -sign(dot({b}, {c}));", vec3tmp())[0]
        of 26:# MULTIPLY_ADD
            vec_out = mkOutput(&"$0 = {a}*{b}+{c};", vec3tmp())[0]
        else:
            raise newException(ValueError, fmt"Unknown operation {operation}")
        if use_clamp:
            vec_out = clamp(vec_out, 0, 1)
            flt_out = clamp(flt_out, 0, 1)
        return @[vec_out, flt_out]
    ,

    "ShaderNodeWavelength": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let wavelength = ins["Wavelength"].flt()
        # TODO
        return mkOutput(&"", vec4tmp())
    ,

    #----- SHADER NODES -----#

    "ShaderNodeEmission": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec3()
        let strength = ins["Strength"].flt()
        # TODO: what is this? it's not even in the UI, it's in all shader nodes
        # let weight = ins["Weight"]
        return mkOutput(&"$0 = vec4({color} * {strength}, 1.0);", vec4tmp())
    ,
    "ShaderNodeBackground": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec4()
        let strength = ins["Strength"].flt()
        return mkOutput(&"$0 = {color} * {strength};", vec4tmp())
    ,
    "ShaderNodeAddShader": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let a = ins["Shader"].vec4()
        let b = ins["Shader1"].vec4()
        return mkOutput(&"$0 = vec4({a}.rgb + {b}.rgb, {a}.a * {b}.a);", vec4tmp())
    ,
    "ShaderNodeMixShader": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let fac = ins["Fac"].flt()
        let a = ins["Shader"].vec4()
        let b = ins["Shader1"].vec4()
        return mkOutput(&"$0 = mix({a}, {b}, {fac});", vec4tmp())
    ,
    "ShaderNodeHoldout": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO: how to distinguish between regular alpha and holdout?
        #       also should we allow mixing them as in blender?
        return @[vec4("vec4(0,0,0,1)")]
    ,
    "ShaderNodeBsdfTransparent": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec3()
        return @[vec4(&"vec4({color},0)")]
    ,
    "ShaderNodeBsdfRefraction": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let color = ins["Color"].vec3()
        let roughness = ins["Roughness"].flt()
        let ior = ins["IOR"].flt()
        let normal = ins["Normal"].normal_socket_to_world
        return mkOutput(&"$0 = refraction_node({color}, " &
            &"{world_incoming()}, " &
            &"{normal}, " &
            &"{world_position()}, " &
            &"{roughness}, " &
            &"{ior} " &
            &");", vec4tmp())
    ,
                             
    #----- VECTOR NODES -----#

    "ShaderNodeDisplacement": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let space = ins.node.custom1.i16[0] # 1: object, 2: world
        let height = ins["Height"].flt()
        let midlevel = ins["Midlevel"].flt()
        let scale = ins["Scale"].flt()
        var normal = ins["Normal"].normal_socket_to_world(normalize=true)
        if space == 1:
           normal = flt &"""{normal} * {uniform(flt("average_world_scale"))}"""
        return mkOutput(&"$0 = {normal}*({height}-{midlevel})*{scale};", vec3tmp())
    ,
    "ShaderNodeBump": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO
        let strength = ins["Strength"].flt()
        let distance = ins["Distance"].flt()
        let height = ins["Height"].flt()
        let normal = ins["Normal"].normal_socket_to_world
        return mkOutput(&"", vec3tmp())
    ,
    "ShaderNodeMapping": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let vtype = ins.node.custom1.i16[0] # point, texture, vector, normal
        let vector = ins["Vector"].vec3()
        let location = ins["Location"].vec3()
        let rotation = ins["Rotation"].vec3()
        let scale = ins["Scale"].vec3()
        let rot_m3 = rotation.euler_xyz_to_mat3
        # TODO: use safe divide?
        return case vtype:
            of 0: # point
                mkOutput(&"$0 = ({rot_m3} * ({vector} * {scale})) + {location};", vec3tmp())
            of 1: # texture
                mkOutput(&"$0 = transpose({rot_m3}) * ({vector} - {location}) / {scale};", vec3tmp())
            of 2: # vector
                mkOutput(&"$0 = {rot_m3} * ({vector} * {scale});", vec3tmp())
            of 3: # normal
                mkOutput(&"$0 = normalize({rot_m3} * ({vector} / {scale}));", vec3tmp())
            else:
                raise newException(ValueError, fmt"Unknown vector type {vtype}")
    ,
    "ShaderNodeNormal": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let normal0 = ins.outputs["Normal"]
        let normal1 = ins["Normal"].vec3()
        return mkOutput(&"$1 = dot({normal0},{normal1});", normal0, flttmp())
    ,
    "ShaderNodeNormalMap": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let props = ins.node.storage
        var space = props.space.i32[0]
        let strength = ins["Strength"].flt()
        var nmap = ins["Color"].vec3().color_to_vectors()
        if space in [3,4]:
            # old blender normals
            # TODO: It has been broken in blender's side for a long time!
            space -= 2
            nmap = vec3(&"vec3({nmap}.x, -{nmap}.y, -{nmap}.z)")
        case space:
            of 0: # tangent
                let uv_map = props.uv_map.str
                let attr_tg = if uv_map != "":
                    varying(Tangent, vec4("tg_" & uv_map.replace(re"[^_A-Za-z0-9]", "")))
                else:
                    varying(Tangent, vec4("tg_active"))
                let normal = view_normal()
                # $1 is the binormal
                var out_normal = mkOutput(dedent &"""
                    $1 = {attr_tg}.w * cross({normal}, {attr_tg}.xyz);
                    $0 = normalize({nmap}.x * {attr_tg}.xyz + {nmap}.y * $1 + {nmap}.z * {normal});
                """, vec3tmp(), vec3tmp())[0]
                out_normal = mix(normal, out_normal, strength).view2world_dir()
                return @[out_normal]
            of 1: # object
                # only object normals seem to have strength squared...
                return  mkOutput(dedent &"""
                    $0 = normalize(({uniform(mat4("model_matrix"))} * vec4({nmap}, 0.0)).xyz);
                    $0 = normalize(mix({world_normal().normalize()}, $0, {strength}*{strength}));
                """, vec3tmp())
            of 2: # world
                return  mkOutput(dedent &"""
                    $0 = normalize({nmap});
                    $0 = normalize(mix({world_normal().normalize()}, $0, {strength}));
                """, vec3tmp())
            else:
                raise newException(ValueError, fmt"Unknown normal map space {space}")
    ,
    "ShaderNodeVectorCurve": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let storage = ins.node.storage
        let flags = storage.flag.i32[0]
        let extrapolate = (flags and 16) != 0
        let fac = ins["Fac"].flt
        let vec = ins["Vector"].vec3
        var curves: seq[seq[CurveMapPoint]]
        var start_x = -0.001
        var end_x = 0.001
        for cm in storage.cm:
            if not cm.curve.valid:
                continue
            var points: seq[CurveMapPoint]
            for p in cm.curve[0 ..< cm.totpoint.i16[0]]:
                let ptype = ((p.flag.i16[0] shr 1) and 3).CurveMapPointHandleType
                points.add CurveMapPoint(vec:[
                    (0.0,0.0),
                    (p.x.f32[0].float,p.y.f32[0].float),
                    (0.0,0.0),
                ], h1: ptype, h2: ptype)
            points = calc_all_handles_curvemap(points)
            # TODO: if a curve has 2 points or less, don't include it in the span
            #       and use the curve equation exclusively
            start_x = min(start_x, points[0].vec[1][0])
            end_x = max(end_x, points[^1].vec[1][0])
            curves.add points
        let width = 256
        # adjust the texture so the center of first and last pixels match 
        # with the extents (so they span two less halfs of width)
        let px_size = (end_x-start_x)/(width-1).float32
        let tex_start_x = start_x - px_size/2
        let tex_end_x = end_x + px_size/2
        let pixels = curves_to_mapping_float_pixels(curves, width, tex_start_x, tex_end_x, 1.0, 0.0)

        # TODO: use R_f32
        let s = texture_uniform(
            TexturePixels(pixels: pixels, width: width, height:1, format: RGBA_f32, tex_type: Tex2D)
        )

        let in_scale = 1/(tex_end_x-tex_start_x)
        let in_offset = -tex_start_x * in_scale

        if not extrapolate:
            return mkOutput(dedent &"""
                $0 = vec3(
                    texture({s}, vec2({vec}.x*{in_scale}+{in_offset}, 0.5)).x,
                    texture({s}, vec2({vec}.y*{in_scale}+{in_offset}, 0.5)).y,
                    texture({s}, vec2({vec}.z*{in_scale}+{in_offset}, 0.5)).z
                );
                $0 = mix({vec}, $0, {fac});
            """, vec3tmp())
        
        let eq0 = get_line_equations(curves[0][0], curves[0][^1])
        let eq1 = get_line_equations(curves[1][0], curves[1][^1])
        let eq2 = get_line_equations(curves[2][0], curves[2][^1])
        # m and b are the variables of a linear equation (y = mx + b)
        # m is the slope and b is y-intercept
        let left_m = vec3(eq0[0][0], eq1[0][0], eq2[0][0])
        let left_b = vec3(eq0[0][1], eq1[0][1], eq2[0][1])
        let right_m = vec3(eq0[1][0], eq1[1][0], eq2[1][0])
        let right_b = vec3(eq0[1][1], eq1[1][1], eq2[1][1])
        
        return mkOutput(dedent &"""
            $0 = vec3(
                {vec}.x <= {start_x} ? {vec}.x*{left_m.x }+{left_b.x} :
                {vec}.x >= {end_x} ? {vec}.x*{right_m.x}+{right_b.x} :
                    texture({s}, vec2({vec}.x*{in_scale}+{in_offset}, 0.5)).x,
                {vec}.y <= {start_x} ? {vec}.y*{left_m.y }+{left_b.y} :
                {vec}.y >= {end_x} ? {vec}.y*{right_m.y}+{right_b.y} :
                    texture({s}, vec2({vec}.y*{in_scale}+{in_offset}, 0.5)).y,
                {vec}.z <= {start_x} ? {vec}.z*{left_m.z }+{left_b.z} :
                {vec}.z >= {end_x} ? {vec}.z*{right_m.z}+{right_b.z} :
                    texture({s}, vec2({vec}.z*{in_scale}+{in_offset}, 0.5)).z
            );
            $0 = mix({vec}, $0, {fac});
        """, vec3tmp())
    ,
    "ShaderNodeVectorDisplacement": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO
        let vector = ins["Vector"].vec3_or_4()
        let midlevel = ins["Midlevel"].flt()
        let scale = ins["Scale"].flt()
        return mkOutput(&"", vec3tmp())
    ,
    "ShaderNodeVectorRotate": proc(ins: InputMapper): seq[Expr] {.closure.} =
        # TODO
        let vector = ins["Vector"].vec3()
        let center1 = ins["Center"].vec3()
        let axis = ins["Axis"].vec3()
        let angle = ins["Angle"].flt()
        let rotation = ins["Rotation"].vec3()
        return mkOutput(&"", vec3tmp())
    ,
    "ShaderNodeVectorTransform": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let storage = ins.node.storage
        # 0 vector, 1 point, 2 normal
        let ttype = storage["type"].i32[0]
        # 0 world, 1 object, 2 camera
        let convert_from = storage.convert_from.i32[0]
        let convert_to = storage.convert_to.i32[0]
        let vector = ins["Vector"].vec3()
        let combo = (convert_from shl 4) or convert_to
        let mat = case combo:
            of 0x10: uniform(mat4("model_matrix"))
            of 0x01: uniform(mat4("model_matrix_inverse"))
            of 0x02: uniform(mat4("view_matrix"))
            of 0x20: uniform(mat4("view_matrix_inverse"))
            of 0x12: uniform(mat4("model_view_matrix"))
            of 0x21: uniform(mat4("model_view_matrix_inverse"))
            else:
                # Both are the same, don't do anything
                var v_out = vector
                if ttype == 2: # normal
                    v_out = v_out.normalize()
                return @[v_out]
        
        var trans_in = vector
        if convert_from == 2: # camera
            trans_in = mkOutput(&"$0 = vec3({trans_in}.xy, -{trans_in}.z);", vec3tmp())[0]
        
        let w = if ttype == 1: 1.0 # point
        else: 0.0 # direction

        var v_out = mkOutput(&"$0 = ({mat} * vec4({trans_in}, {w})).xyz;", vec3tmp(""))[0]

        if convert_to == 2: # camera
            v_out = mkOutput(&"$0 = vec3({v_out}.xy, -{v_out}.z);", vec3tmp())[0]
        if ttype == 2: # normal
            v_out = v_out.normalize()
        
        return @[v_out]
    ,

    #----- GROUP NODES -----#

    "ShaderNodeGroup": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let node_tree = ins.node.id
        let tree_name = node_tree.name.str.strip2
        if tree_name.endswith(".glsl"):
            context.libraries.incl tree_name
            let func_name = tree_name[0 ..< ^5]
            var args: seq[string]
            for inp_name in ins:
                args.add $ins[inp_name]
            let outs = ins.outputs
            var ovars: seq[Expr]
            for oname in outs:
                args.add "$" & $ovars.len
                let o = outs[oname]
                # type    btype    vlen
                # ----    -----    ----
                # float   B_float  1
                # vector  B_float  3
                # color   B_float  4
                # shader  B_float  4
                # bool    B_char   1    TODO
                # integer B_int32  1    TODO
                ovars.add case o.vlen:
                    of 1: flttmp()
                    of 3: vec3tmp()
                    of 4: vec4tmp()
                    else:
                        assert false, "Unexpected socket type"
                        Expr()
            return mkOutput(&"""{func_name}({args.join ", "});""", ovars)
        else:
            let node_tree = get_node_tree(tree_name)
            context.group_stack.add context.current_group
            context.current_group = ins
            var output_ins: InputMapper
            var node = node_tree.nodes.first
            while node.valid:
                if node.idname.str == "NodeGroupOutput":
                    output_ins = node.toInputMapper
                    break
                node = node.next
            if not output_ins.node.valid:
                return
            for iname in output_ins:
                result.add output_ins[iname]
            context.current_group = context.group_stack.pop()
        return
    ,

    "NodeGroupInput": proc(ins: InputMapper): seq[Expr] {.closure.} =
        let current_group = context.current_group
        context.current_group = context.group_stack.pop()

        for iname in current_group:
            result.add current_group[iname]

        context.group_stack.add context.current_group
        context.current_group = current_group
}.toTable

