# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.


import std/math
import std/fenv

func cubic_bezier*[T: SomeFloat](t, p0, p1, p2, p3: T): T =
    let t2 = t * t
    let t3 = t2 * t
    let c0 = p0
    let c1 = -3.0 * p0 + 3.0 * p1
    let c2 = 3.0 * p0 - 6.0 * p1 + 3.0 * p2
    let c3 = -p0 + 3.0 * p1 - 3.0 * p2 + p3
    return c0 + t * c1 + t2 * c2 + t3 * c3

proc cubic_root[T: SomeFloat](d: T): T = pow(d.abs,1/3).copySign(d)
proc within_0_1[T: SomeFloat](v: T): bool = v >= -T.epsilon and v < 1+T.epsilon

func solve_cubic_roots*[T: SomeFloat](x, p0, p1, p2, p3: T): T =
    # Adapted from Graphics Gems
    # And from Graphics Gems IV
    # Roots3And4.c, solver.c
    # by Jochen Schwarze

    # Instead of giving all solutions, it gives the first that is
    # within 0 and 1, allowing for a bit of imprecision

    const tPI: T = (4 * arctan(1.0))/3

    var s: T

    let c0: T = p0 - x
    let c1: T = -3.0 * p0 + 3.0 * p1
    let c2: T = 3.0 * p0 - 6.0 * p1 + 3.0 * p2
    let c3: T = -p0 + 3.0 * p1 - 3.0 * p2 + p3

    if abs(c3) <= T.epsilon:
        if abs(c1) > T.epsilon:
                s = -c0 / c1
        if within_0_1(s):
                return 1
        if abs(c0) <= T.epsilon:
                return 1
        return 0

    # normal form:
    # x^3 + Ax^2 + Bx + C = 0

    let a: T = c2 / c3
    let b: T = c1 / c3
    let c: T = c0 / c3

    # substitute x = y - a/3 to eliminate quadric term:
    # x^3 +px + q = 0

    let ao3 = a/3
    let aa = a * a

    let p: T = (1/3) * (-(1/3) * aa + b)
    let q: T = 0.5 * (2/27 * a * aa - (1/3) * a * b + c)

    # use Cardano's formula
    let cp: T = p * p * p
    let d: T = q * q + cp

    if abs(d) <= T.epsilon:
        if abs(q) <= T.epsilon:
            # one triple solution
            s = 0.0
            # console.log 'triple'
            return s
        else:
            let u: T = cubic_root(-q)

            # one single and one double solution
            s = 2.0 * u
            # try next
            if not within_0_1(s - ao3):
                s = -u
    elif d < 0.0:
        # three real solutions
        let phi: T = (1/3) * arccos(-q / sqrt(-cp))
        let t: T = 2.0 * sqrt -p

        s = t * cos phi
        if not within_0_1(s - ao3):
            # try next
            s = -t *  cos(phi + tPI)
            if not within_0_1(s - ao3):
                s = -t * cos(phi - tPI)
    else:
        # one real solution
        let S: T = sqrt d
        let u: T = cubic_root(S - q)
        let v: T = -cubic_root(S + q)
        s = u + v

    # resubstitute
    s -= ao3
    return s
