# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import ../types
import std/options
import vmath except Quat
import ../quat
import ../util

when defined(nimdoc):
    type TYPES* = CameraType | SensorFit | Camera

# Forward declarations
proc update_projection*(self: Camera)
proc calculate_projection*(self: Camera)
proc calculate_culling_planes*(self: Camera)
proc get_oblique_projection_matrix_into*(self: Camera, clip_plane: Vec4): Mat4
# proc ensure_shadow_map_count*(self: Camera)
# End forward declarations

import ./gameobject
# import ../graphics/ubo

method is_camera*(self: GameObject): bool {.base.} =
    return false
method get_camera*(self: GameObject): Camera {.base.} =
    return nil
method is_camera*(self: Camera): bool =
    return true
method get_camera*(self: Camera): Camera =
    return self

proc newCamera*(engine: MyouEngine, name: string="camera", scene: Scene = nil,
        near_plane: float32 = 0.1,
        far_plane: float32 = 10000,
        field_of_view: float32 = toRadians(30),
        ortho_scale: float32 = 8,
        aspect_ratio: float32 = 1,
        cam_type: CameraType = Perspective,
        sensor_fit: SensorFit = Auto,
        shift: Vec2 = vec2(),
    ): Camera =
    var self = new Camera
    discard procCall(self.GameObject.initGameObject(engine, name))
    self.otype = TCamera
    self.near_plane = near_plane
    self.far_plane = far_plane
    self.field_of_view = field_of_view
    self.ortho_scale = ortho_scale
    self.aspect_ratio = aspect_ratio
    self.cam_type = cam_type
    self.sensor_fit = sensor_fit
    self.shift = shift
    self.projection_matrix = mat4()
    self.projection_matrix_inverse = mat4()
    self.world_to_screen_matrix = mat4()
    # self.camera_render_ubo = self.engine.newUBO("CameraRenderUniform", CameraRenderUniform, 1)
    # self.shadow_maps_ubo = self.engine.newUBO("ShadowMapUniform", ShadowMapUniform, 0)
    self.update_projection()
    return self

proc clone*(self: Camera, 
        recursive: bool=true,
        with_behaviours: bool=true,
        name: string=self.name,
        new_parent: GameObject=self.parent,
        scene: Scene=self.scene,
        instance_body: bool=true
    ): Camera =
    var n = new Camera
    n[] = self[]
    discard procCall(self.GameObject.clone_impl(n.GameObject, recursive, with_behaviours, name, new_parent, scene, instance_body))
    return n

proc instance_physics*(self: Camera) =
    discard

proc get_ray_direction*(self: Camera, x, y: float32): Vec3 =
    return self.get_world_rotation * (self.projection_matrix_inverse * 
        vec3(x * 2 - 1, 1 - y * 2, 1))

proc get_ray_direction_local*(self: Camera, x, y: float32): Vec3 =
    assert self.rotation_order == Quaternion
    return self.rotation * (self.projection_matrix_inverse * 
        vec3(x * 2 - 1, 1 - y * 2, 1))

# proc get_position_at_depth*(self: camera, x, y: float32): Vec3 =
#     return vec3(x,y)

# proc look_at*(self: Camera, target: unknown, options: Table[string, unknown] = {:}.toTable) =
#     options.front ?= "-Z"
#     options.up ?= "+Y"
#     procCall(self.GameObject.look_at(target, options))

proc is_vertical_fit*(self: Camera): bool =
    return case self.sensor_fit:
        of Auto: self.aspect_ratio <= 1
        of Horizontal: false
        of Vertical: true
        of Cover: self.aspect_ratio <= self.target_aspect_ratio
        of Contain: self.aspect_ratio > self.target_aspect_ratio

proc set_projection_matrix*(self: Camera, matrix: Mat4, adjust_aspect_ratio: bool = false) =
    self.projection_matrix = matrix
    if adjust_aspect_ratio:
        let m_ratio = matrix[0, 0] / matrix[1, 1]
        # TODO: Use sensor_fit, we'll assume 'cover' for now
        # TODO: it's expecting an inverse aspect ratio, probably
        let ratio_factor = m_ratio / self.aspect_ratio
        self.projection_matrix = if ratio_factor > 1:
            scale(vec3(1, ratio_factor, 1)) * self.projection_matrix
        else:
            scale(vec3(1 / ratio_factor, 1, 1)) * self.projection_matrix
    self.projection_matrix_inverse = inverse(self.projection_matrix)
    self.calculate_culling_planes()

proc update_projection*(self: Camera) =
    self.calculate_projection()
    self.calculate_culling_planes()

proc calculate_projection*(self: Camera) =
    let near_plane = self.near_plane
    let far_plane = self.far_plane
    var top, right, bottom, left: float32
    if not self.fov_4.isSome:
        let half_size = case self.cam_type:
            of Perspective:
                near_plane * tan(self.field_of_view / 2)
            of Orthographic:
                self.ortho_scale / 2
        if self.is_vertical_fit():
            top = half_size
            if self.sensor_fit == Contain or self.sensor_fit == Cover:
                top /= self.target_aspect_ratio
            right = top * self.aspect_ratio
        else:
            right = half_size
            top = right / self.aspect_ratio
        bottom = -top
        left = -right
    else:
        # Custom FoV in each direction, for VR
        let (t,r,b,l) = self.fov_4.get
        top = near_plane * tan(t)
        right = near_plane * tan(r)
        bottom = near_plane * tan(-b)
        left = near_plane * tan(-l)
    let np_width = right - left
    let np_height = top - bottom
    left += self.shift.x * np_width
    right += self.shift.x * np_width
    top += self.shift.y * np_height
    bottom += self.shift.y * np_height
    let a = (right + left) / np_width
    let b = (top + bottom) / np_height
    let c = -(far_plane + near_plane) / (far_plane - near_plane)
    if self.cam_type == Perspective:
        let d = -(2 * far_plane * near_plane) / (far_plane - near_plane)
        let x = (2 * near_plane) / np_width
        let y = (2 * near_plane) / np_height
        # Column major order (each row is actually a column)
        self.projection_matrix = mat4(
            x, 0, 0, 0,
            0, y, 0, 0,
            a, b, c, -1,
            0, 0, d, 0
        )
        self.projection_matrix_inverse = inverse(self.projection_matrix)
    else:
        let d = -2 / (far_plane - near_plane)
        let x = 2 / np_width
        let y = 2 / np_height
        self.projection_matrix = mat4(
            x, 0, 0, 0,
            0, y, 0, 0,
            0, 0, d, 0,
            -a,-b,c, 1
        )
    self.projection_matrix_inverse = inverse(self.projection_matrix)

proc calculate_culling_planes*(self: Camera) =
    self.cull_planes = self.projection_matrix.get_culling_planes()

proc get_oblique_projection_matrix_into*(self: Camera, clip_plane: Vec4): Mat4 =
    result = self.projection_matrix
    var q2 = vec4(clip_plane.x.sign, clip_plane.y.sign, 1, 1)
    q2 = self.projection_matrix_inverse * q2
    let d = 2 / dot(clip_plane, q2)
    let c = clip_plane * d
    # last row is always (0,0,-1,0)... NOTE: maybe not in vulkan?
    result[0, 2] = c.x - result[0, 3]
    result[1, 2] = c.y - result[1, 3]
    result[2, 2] = c.z - result[2, 3]
    result[3, 2] = c.w - result[3, 3]

# proc ensure_shadow_map_count*(self: Camera) =
#     let count = self.scene.max_shadow_maps
#     if self.shadow_maps_ubo.storage(ShadowMapUniform).len < count:
#         self.shadow_maps_ubo.resize(ShadowMapUniform, count)
