# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import ./types

# Forward declarations
proc newScreen*(engine: MyouEngine, width, height: int32, title: string): Screen
proc destroy*(self: Screen)
proc resize*(self: Screen, width, height: int32, orientation = self.orientation)
proc add_viewport*(self: Screen, camera: Camera)
proc clear_all_callbacks*(self: Screen)
proc emulateMouseWithTouch*(screen: Screen, touch: int32, ending: bool, x, y: float32)
# End forward declarations

import std/bitops
import std/sequtils
import std/math
import vmath except Quat
import ./graphics/framebuffer
import ./objects/camera
import ./platform/platform
import ./input
import ./util

proc newScreen*(engine: MyouEngine, width, height: int32, title: string): Screen =
    result = new Screen
    result.engine = engine
    result.width = width
    result.height = height
    # note: the window will overwrite width and height in some platforms
    let window = make_window(width, height, title)
    window.screen = result
    result.window = cast[pointer](window)
    result.framebuffer = engine.newMainFramebuffer()
    result.enabled = true
    engine.screens.add result
    if engine.screen == nil:
        engine.screen = result
    result.resize(result.width, result.height)
    result.pre_draw = proc(self: Screen) = discard
    result.post_draw = proc(self: Screen) = discard
    result.frame_interval = 1

proc destroy*(self: Screen) =
    self.engine.screens.remove self
    if self.engine.screen == self and self.engine.screens.len != 0:
        self.engine.screen = self.engine.screens[0]

proc resize*(self: Screen, width, height: int32, orientation = self.orientation) =
    self.width = width
    self.height = height
    self.orientation = orientation
    for vp in self.viewports:
        let (x,y,w,h) = vp.rect
        # TODO: verify this doesn't leave gaps/overlaps
        vp.rect_pix = (
            round(x*width.float32).int32,
            round(y*height.float32).int32,
            round(w*width.float32).int32,
            round(h*height.float32).int32,
        )
        if height != 0:
            vp.camera.aspect_ratio = width / height
        vp.camera.update_projection()
    for f in self.resize_callbacks:
        f(self)

proc add_viewport*(self: Screen, camera: Camera) =
    let vp = new Viewport
    vp.camera = camera
    vp.clear_color = true
    vp.clear_depth = true
    vp.rect = (0'f32, 0'f32, 1'f32, 1'f32)
    self.viewports.add vp
    self.resize(self.width, self.height)

proc `vsync=`*(self: Screen, vsync: bool) =
    cast[Window](self.window).set_vsync vsync

proc get_ray_direction*(viewport: Viewport, position: Vec2): Vec3 =
    let x = (position.x - viewport.rect_pix[0].float32) / viewport.rect_pix[2].float32
    let y = (position.y - viewport.rect_pix[1].float32) / viewport.rect_pix[3].float32
    return viewport.camera.get_ray_direction(x,y)

proc get_ray_direction_local*(viewport: Viewport, position: Vec2): Vec3 =
    let x = (position.x - viewport.rect_pix[0].float32) / viewport.rect_pix[2].float32
    let y = (position.y - viewport.rect_pix[1].float32) / viewport.rect_pix[3].float32
    return viewport.camera.get_ray_direction_local(x,y)

proc get_pixels_at_depth*(viewport: Viewport, depth: float32): float32 =
    ## Returns the length of one world unit (usually one meter) in screen pixels,
    ## at a given depth.
    
    # TODO: shift_x
    let p = vec4(1.0, 0.0, -depth, 1.0)
    let s = viewport.camera.projection_matrix * p
    return viewport.rect_pix[2].float32 * 0.5 * s.x/s.w

proc clear_all_callbacks*(self: Screen) =
    self.key_callbacks.setLen 0
    self.mouse_button_callbacks.setLen 0
    self.mouse_move_callbacks.setLen 0

proc size*(self: Screen): IVec2 {.inline.} = ivec2(self.width, self.height)


proc emulateMouseWithTouch*(screen: Screen, touch: int32, ending: bool, x, y: float32) =
    template send_move(x, y: untyped) =
        let buttons = screen.last_buttons
        let e = MouseMoveEvent(
                left: (buttons and 1).bool,
                middle: (buttons and 2).bool,
                right: (buttons and 4).bool,
                # shiftKey: shiftKey, ctrlKey: ctrlKey, altKey: altKey, metaKey: metaKey,
                position: vec2(x,y),
                movement: vec2(x-screen.last_x, y-screen.last_y),
            )
        for cb in screen.mouse_move_callbacks:
            cb(e)
            if screen.break_current_callbacks:
                screen.break_current_callbacks = false
                break
        screen.last_x = x
        screen.last_y = y
    template send_btn(pressed1, btn, x, y: untyped) =
        let e = MouseButtonEvent(
            pressed: pressed1,
            button: cast[MouseButton](btn),
            position: vec2(x,y)
        )
        for cb in screen.mouse_button_callbacks:
            cb(e)
            if screen.break_current_callbacks:
                screen.break_current_callbacks = false
                break
        if pressed1:
            screen.last_buttons = screen.last_buttons or (1'i8 shl btn)
        else:
            screen.last_buttons = screen.last_buttons and not (1'i8 shl btn)
        screen.last_x = x
        screen.last_y = y
    template send_wheel(x,y: untyped) =
        let e = MouseWheelEvent(
            movement: vec2(x,y),
        )
        for cb in screen.mouse_wheel_callbacks:
            cb(e)
            if screen.break_current_callbacks:
                screen.break_current_callbacks = false
                break
    
    let prev_count = screen.touches.len
    var index = -1
    for i,t in screen.touches.mpairs:
        if t[0] == touch:
            index = i
            t[1] = vec2(x,y)
            break
    if not ending and index != -1: # move
        if index == screen.touches.high:
            # when we receive the last touch, we received them all, time to
            # calculate mousemove and scroll wheel
            var delta: Vec2
            for i,(_,v) in screen.touches:
                delta += v - screen.prev_touches[i][1]
            if delta.lengthSq != 0.0:
                delta /= screen.touches.len.float32
                send_move(screen.last_x + delta.x, screen.last_y + delta.y)
            # when we have more than one touch we calculate zoom and angle
            # from a pinch gesture
            if index != 0:
                let v = screen.touches[0][1] - 
                        screen.touches[^1][1]
                let v0 = screen.prev_touches[0][1] -
                        screen.prev_touches[^1][1]
                let zoom = v0.length/v.length
                let angle = fixAngle(v.angle - v0.angle)
                if zoom != 1.0 or angle != 0.0:
                    send_wheel(angle, -log(zoom, 1.2))
            # finally, store the current touches to use next time
            screen.prev_touches = screen.touches
    elif not ending: # start
        screen.touches.add (touch.int32, vec2(x,y))
        let button = prev_count
        if screen.last_buttons == 0:
            send_btn(true, button, x, y)
        else:
            let previous = screen.last_buttons.countTrailingZeroBits
            if previous < button:
                send_btn(true, button, screen.last_x, screen.last_y)
                send_btn(false, previous, screen.last_x, screen.last_y)
        screen.prev_touches = screen.touches
    else: # end
        screen.touches.keepItIf it[0] != touch
        if screen.touches.len == 0:
            let previous = screen.last_buttons.countTrailingZeroBits
            send_btn(false, previous, screen.last_x, screen.last_y)
        screen.prev_touches = screen.touches

proc switch*(self: Screen): bool {.discardable.} = 
    self.platform_switch_screen()

