/*

    Sources:
        https://github.com/SaschaWillems/Vulkan
        https://seblagarde.files.wordpress.com/2015/07/course_notes_moving_frostbite_to_pbr_v32.pdf
        https://cdn2.unrealengine.com/Resources/files/2013SiggraphPresentationsNotes-26915738.pdf
        https://seblagarde.wordpress.com/2012/09/29/image-based-lighting-approaches-and-parallax-corrected-cubemap/

*/

// Normal Distribution function --------------------------------------
// float D_GGX(float dotNH, float roughness){
//     float alpha = roughness * roughness;
//     float alpha2 = alpha * alpha;
//     float denom = dotNH * dotNH * (alpha2 - 1.0) + 1.0;
//     return (alpha2)/(PI * denom*denom); 
// }

// Geometric Shadowing function --------------------------------------
float G_SchlicksmithGGX(float dotNL, float dotNI, float roughness){
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;
    float GL = dotNL / (dotNL * (1.0 - k) + k);
    float GV = dotNI / (dotNI * (1.0 - k) + k);
    return GL * GV;
}

// Fresnel function ----------------------------------------------------
vec3 F_Schlick(float cosTheta, float metallic, vec3 base, float ior){
    // TODO: supply f0 to ensure it's optimized out when constant?
    float f0 = pow((1.0 - ior) / (1.0 + ior), 2.0);
    vec3 F0 = mix(vec3(f0), base, metallic);
    // cosTheta can be 1, so we avoid nan by adding epsilon
    // vec3 F = F0 + (1.0 - F0) * pow(1.000000119 - cosTheta, 5.0); 
    vec3 F = F0 + (1.0 - F0) * pow(max(0.0, 1.0 - cosTheta), 5.0); 
    return F;
}


// Diffuse point light

float sphere_attenuation(float dist_sq, float inv_squared_cutoff){
    float f = max(0.0, 1.0-(dist_sq * inv_squared_cutoff));
    return f*f;
}

float diffuse_point_falloff(float dist_sq, float power, float radius_sq, float inv_squared_cutoff){
    // NOTE: without horizon handling, changes in radius doesn't make sense,
    // so we can probably just ignore it altogether
    float final_dist_sq = dist_sq;
    // float final_dist_sq = max(dist_sq, radius_sq);
    float att = sphere_attenuation(final_dist_sq, inv_squared_cutoff);
    float fall = 1.0/final_dist_sq;
    // constant found by trial and error to match eevee
    return att * fall * power * 0.001 * 40.49;
}

// Specular point light
// Not correct, but good enough for now

float specular_point_falloff(float dist, float dist_sq, float power, float radius_sq, float inv_squared_cutoff){
    float final_dist_sq = max(dist_sq, radius_sq);
    float att = sphere_attenuation(final_dist_sq, inv_squared_cutoff);
    // constants found by trial and error to TRY to match eevee
    float fall = 1.0/(dist*1.222 + final_dist_sq);
    return att * fall * power * 0.001 * PI * 40.49;
}

// diffuse sun
// TODO: radius
float diffuse_sun_light(float power){
    return 0.2518836 * power;
}

// float specular_sun_light(float power){
//     return 1.0 * power;
// }

#ifndef FIX_BORDERS
#define FIX_BORDERS 1
#endif

vec3 sample_environment(vec3 v, vec3 pos, float roughness){
    // TODO TODO: SKIP UNUSED UBO SLOTS!!
    #if MAX_CUBE_MAPS
    vec3 color = vec3(0);
    for(int i=0;i<MAX_CUBE_MAPS;i++){
        vec3 cube_pos = (cube_map_infos[i].world2cube * vec4(pos, 1.0)).xyz;
        float dist;
        if(cube_map_infos[i].influence_type == 0.0){
            dist = dot(cube_pos,cube_pos);
            if(dist > 1.0) continue;
            dist = sqrt(dist);
        }else{
            vec3 box = abs(cube_pos);
            dist = max(max(box.x, box.y), box.z);
            if(dist >= 1.0) continue;
        }
        float infl = (1.0 - dist) * cube_map_infos[i].falloff_inv;

        vec3 V = normalize((cube_map_infos[i].world2cube * vec4(v, 0.0)).xyz);
        // I don't know why this is nan sometimes
        if(V != V) continue;

        // Parallax correction
        // TODO: have option for baking these with #defines
        if(cube_map_infos[i].parallax_type != -1.0){
            float vdist;
            cube_pos *= cube_map_infos[i].par_dist_rel_inv;
            if(cube_map_infos[i].parallax_type == 1.0){
                vec3 plane0  = (vec3( 1.0) - cube_pos) / V;
                vec3 plane1 = (vec3(-1.0) - cube_pos) / V;
                vec3 furthest = max(plane0, plane1);
                vdist = min(furthest.x, min(furthest.y, furthest.z));
            }else{
                /* ray sphere intersection */
                float a = dot(V, V);
                float b = dot(V, cube_pos);
                float c = dot(cube_pos, cube_pos) - 1.0;

                vdist = 1.0e15;
                float determinant = b * b - a * c;
                if (determinant >= 0.0)
                    vdist = (sqrt(determinant) - b) / a;
            }
            vec3 intersection = cube_pos + V * vdist;
            V = normalize(intersection);
        }
#ifdef PREFILTERED_CUBEMAPS
        // vec3 e = textureLod(cube_maps[i], V, roughness * cube_map_infos[i].roughness_lod).rgb
        //         * cube_map_infos[i].intensity;
        vec3 e;
        if(i==0) e = textureLod(cube_maps[0], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 2
        else if(i==1) e = textureLod(cube_maps[1], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 3
        else if(i==2) e = textureLod(cube_maps[2], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 4
        else if(i==3) e = textureLod(cube_maps[3], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 5
        else if(i==4) e = textureLod(cube_maps[4], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 6
        else if(i==5) e = textureLod(cube_maps[5], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 7
        else if(i==6) e = textureLod(cube_maps[6], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #if MAX_CUBE_MAPS >= 8
        else if(i==7) e = textureLod(cube_maps[7], V, roughness * cube_map_infos[i].roughness_lod).rgb * cube_map_infos[i].intensity;
        #endif
        #endif
        #endif
        #endif
        #endif
        #endif
        #endif
#else
        vec3 e = PrefilterEnvMap(cube_maps[i], roughness, V)
                * cube_map_infos[i].intensity;
#endif
        color = mix(color, e, min(infl, 1.0));
    }
    return color;
    #else
    return vec3(0);
    #endif
}

vec4 diffuse_node(vec3 base, vec3 I, vec3 normal, vec3 pos){
    vec3 N = normal;
#if FIX_BORDERS
    float dotNI = dot(N, I);
    if(dotNI < 0.001){
        N = normalize(N - I * dotNI);
    }
#endif
    vec3 diffuse = base;
    vec3 color = vec3(0);
    #if MAX_POINT_LIGHTS
    for(int i=0;i<MAX_POINT_LIGHTS;i++){
        PointLight ld = point_lights[i];
        vec3 p = ld.pos.xyz - pos;
        vec3 L = normalize(p);
        float dotNL = clamp(dot(N, L), 0.0, 1.0);
        if(dotNL <= 0.0) continue;
        float l_dist_sq = dot(p,p);
        float l_dist = sqrt(l_dist_sq);
        vec3 dif = base * diffuse_point_falloff(l_dist_sq, 
            ld.light_diffuse_power,
            ld.light_radius_sq,
            ld.light_inv_squared_cutoff);
        color += dotNL * dif * ld.color.rgb;
    }
    #endif
    #if MAX_SUN_LIGHTS
    for(int i=0;i<MAX_SUN_LIGHTS;i++){
        SunLight ld = sun_lights[i];
        vec3 L = ld.light_direction;
        float dotNL = clamp(dot(N, L), 0.0, 1.0);
        if(dotNL <= 0.0) continue;
        vec3 dif = base * diffuse_sun_light(ld.light_diffuse_power);
        color += dotNL * dif * ld.color.rgb;
    }
    #endif
    // vec3 env_diffuse = sample_environment(N, pos, 1.0) * 0.1;
    // color += mix(env_diffuse, env_specular, F);
    return vec4(color, 1.0);
}

vec4 principled_node(vec3 base, vec3 I, vec3 normal, vec3 pos, float metallic, float roughness, float ior){
    vec3 N = normal;
#if FIX_BORDERS
    float dotNI = dot(N, I);
    if(dotNI < 0.001){
        N = normalize(N - I * dotNI);
        dotNI = 0.001;
    }
#else
    float dotNI = abs(dot(N, I));
#endif
    float rroughness = max(0.05, roughness);
    // F = Fresnel factor (Reflectance depending on angle of incidence)
    vec3 F = F_Schlick(dotNI, metallic, base, ior);
    // shared terms with all specular lights
    vec3 spec_common = mix(vec3(1), base, metallic * 0.5)
        * F / (4.0 * dotNI);
    // shared terms with all diffuse lights
    vec3 diffuse = base * (1.0 - metallic) * (1.0 - F*(1.0-roughness));
    vec3 color = vec3(0);
    #if MAX_POINT_LIGHTS
    for(int i=0;i<MAX_POINT_LIGHTS;i++){
        PointLight ld = point_lights[i];
        vec3 p = ld.pos.xyz - pos;
        vec3 L = normalize(p);
        float dotNL = clamp(dot(N, L), 0.0, 1.0);
        if(dotNL <= 0.0) continue;
        vec3 H = normalize (I + L);
        float dotNH = clamp(dot(N, H), 0.0, 1.0);
        float l_dist_sq = dot(p,p);
        float l_dist = sqrt(l_dist_sq);

        // D = Normal distribution (Distribution of the microfacets)
        float D = D_GGX(dotNH, roughness); 
        // G = Geometric shadowing term (Microfacets shadowing)
        float G = G_SchlicksmithGGX(dotNL, dotNI, rroughness);

        // vec3 spec = D * F * G / (4.0 * dotNL * dotNI);
        // spec *= mix(vec3(1), base, metallic * 0.5);
        // == the two lines above equal the one below ==
        vec3 spec = D * G * spec_common / dotNL;

        spec *= specular_point_falloff(l_dist, l_dist_sq,
            ld.light_specular_power,
            ld.light_radius_sq,
            ld.light_inv_squared_cutoff);
        
        spec = max(vec3(0.0), spec);

        vec3 dif = diffuse * diffuse_point_falloff(l_dist_sq, 
            ld.light_diffuse_power,
            ld.light_radius_sq,
            ld.light_inv_squared_cutoff);
        
        color += dotNL * (dif + spec) * ld.color.rgb;
    }
    #endif
    #if MAX_SUN_LIGHTS
    for(int i=0;i<MAX_SUN_LIGHTS;i++){
        SunLight ld = sun_lights[i];
        vec3 L = ld.light_direction;
        float dotNL = clamp(dot(N, L), 0.0, 1.0);
        if(dotNL <= 0.0) continue;
        vec3 H = normalize (I + L);
        float dotNH = clamp(dot(N, H), 0.0, 1.0);
        float D = D_GGX(dotNH, roughness); 
        float G = G_SchlicksmithGGX(dotNL, dotNI, rroughness);
        vec3 spec = D * G * spec_common / dotNL;
        spec *= ld.light_specular_power;
        vec3 dif = diffuse * diffuse_sun_light(ld.light_diffuse_power);
        float shadow = 1.0;
        if(ld.light_shadow_index >= 0.0){
            shadow = get_shadow(pos, normal, ld.light_shadow_index);
        }
        color += dotNL * (dif + spec) * ld.color.rgb * shadow;
    }
    #endif
    // NOTE TODO: these factors are fake, I'm still figuring out how to make
    // these look good...
    vec3 env_specular = sample_environment(reflect(-I,N), pos, roughness) * (1.0 - roughness);
    #if MAX_SPHERICAL_HARMONICS
        vec3 env_diffuse = get_sh9(N, SH9_coefficients[0]) * 0.3;
    #else
        vec3 env_diffuse = sample_environment(N, pos, 1.0) * 0.1;
    #endif
    // color += mix(env_diffuse, env_specular, F);
    // adding instead of mixing because F is already in diffuse
    color += diffuse * env_diffuse + F * env_specular;
    return vec4(color, 1.0);
}

vec3 refraction_dominant_dir(vec3 N, vec3 V, float roughness, float ior)
{
  /* TODO: This a bad approximation. Better approximation should fit
   * the refracted vector and roughness into the best prefiltered reflection
   * lobe. */
  /* Correct the IOR for ior < 1.0 to not see the abrupt delimitation or the TIR */
  ior = (ior < 1.0) ? mix(ior, 1.0, roughness) : ior;
  float eta = 1.0 / ior;

  float NV = dot(N, -V);

  /* Custom Refraction. */
  float k = 1.0 - eta * eta * (1.0 - NV * NV);
  k = max(0.0, k); /* Only this changes. */
  vec3 R = eta * -V - (eta * NV + sqrt(k)) * N;

  return R;
}


vec4 refraction_node(vec3 base, vec3 I, vec3 normal, vec3 pos, float roughness, float ior){
    vec3 v = refraction_dominant_dir(normal, I, roughness, ior);
    return vec4(base * sample_environment(v, pos, pow(roughness, 3.0)), 1.0);
}

float fresnel_dielectric(vec3 Incoming, vec3 Normal, float ior){
    /* compute fresnel reflectance without explicitly computing
     * the refracted direction */
    float c = abs(dot(Incoming, Normal));
    float g = ior * ior - 1.0 + c * c;
    float result;

    if(g > 0.0) {
        g = sqrt(g);
        float A =(g - c)/(g + c);
        float B =(c *(g + c)- 1.0)/(c *(g - c)+ 1.0);
        result = 0.5 * A * A *(1.0 + B * B);
    }
    else {
        result = 1.0;  /* TIR (no refracted component) */
    }

    return result;
}
